"""Stats-utility functions for ArviZ."""

import logging
from collections.abc import Sequence

import numpy as np
from scipy.interpolate import CubicSpline

__all__ = ["make_ufunc"]

_log = logging.getLogger(__name__)


def make_ufunc(func, n_dims=2, n_output=1, n_input=1, index=Ellipsis, ravel=True, check_shape=None):
    """Make ufunc from a function taking 1D array input.

    Parameters
    ----------
    func : callable
    n_dims : int, optional
        Number of core dimensions not broadcasted. Dimensions are skipped from the end.
        At minimum n_dims > 0.
    n_output : int, optional
        Select number of results returned by `func`.
        If n_output > 1, ufunc returns a tuple of objects else returns an object.
    n_input : int, optional
        Number of **array** inputs to func, i.e. ``n_input=2`` means that func is called
        with ``func(ary1, ary2, *args, **kwargs)``
    index : int, optional
        Slice ndarray with `index`. Defaults to `Ellipsis`.
    ravel : bool, optional
        If true, ravel the ndarray before calling `func`.
    check_shape: bool, optional
        If false, do not check if the shape of the output is compatible with n_dims and
        n_output. By default, True only for n_input=1. If n_input is larger than 1, the last
        input array is used to check the shape, however, shape checking with multiple inputs
        may not be correct.

    Returns
    -------
    callable
        ufunc wrapper for `func`.
    """
    if n_dims is not None and n_dims < 1:
        raise TypeError("n_dims must be one or higher.")

    if n_input == 1 and check_shape is None:
        check_shape = True
    elif check_shape is None:
        check_shape = False

    def _ufunc(*args, out=None, out_shape=None, shape_from_1st=False, **kwargs):
        """General ufunc for single-output function."""
        arys = args[:n_input]
        if n_dims is None:
            element_shape = arys[0].shape
        else:
            element_shape = arys[0].shape[:-n_dims]
        if shape_from_1st and out_shape is not None:
            raise ValueError("out_shape and shape_from_1st are incompatible")
        if out is None and not shape_from_1st:
            if out_shape is None:
                out = np.empty(element_shape)
            else:
                out = np.empty((*element_shape, *out_shape))
        elif check_shape and not shape_from_1st:
            if out.shape != arys[0].shape[:-n_dims]:
                msg = f"Shape incorrect for `out`: {out.shape}."
                msg += f" Correct shape is {arys[-1].shape[:-n_dims]}"
                raise TypeError(msg)
        for idx in np.ndindex(element_shape):
            arys_idx = [ary[idx].ravel() if ravel else ary[idx] for ary in arys]
            aux = np.asarray(func(*arys_idx, *args[n_input:], **kwargs))[index]
            if idx == ():
                aux = np.squeeze(aux)
            if shape_from_1st:
                out = np.empty((*element_shape, *aux.shape))
                shape_from_1st = False
            out[idx] = aux
        return out

    def _multi_ufunc(*args, out=None, out_shape=None, shape_from_1st=False, **kwargs):
        """General ufunc for multi-output function."""
        arys = args[:n_input]
        element_shape = arys[0].shape[:-n_dims]
        if out is None:
            if out_shape is None:
                out = tuple(np.empty(element_shape) for _ in range(n_output))
            else:
                out = tuple(np.empty((*element_shape, *out_shape[i])) for i in range(n_output))

        elif check_shape:
            raise_error = False
            correct_shape = tuple(element_shape for _ in range(n_output))
            if isinstance(out, tuple):
                out_shape = tuple(item.shape for item in out)
                if out_shape != correct_shape:
                    raise_error = True
            else:
                raise_error = True
                out_shape = "not tuple, type={type(out)}"
            if raise_error:
                msg = f"Shapes incorrect for `out`: {out_shape}."
                msg += f" Correct shapes are {correct_shape}"
                raise TypeError(msg)
        for idx in np.ndindex(element_shape):
            arys_idx = [ary[idx].ravel() if ravel else ary[idx] for ary in arys]
            results = func(*arys_idx, *args[n_input:], **kwargs)
            if shape_from_1st:
                out = tuple(np.empty((*element_shape, *res.shape)) for res in results)
                shape_from_1st = False
            for i, res in enumerate(results):
                if idx == ():
                    res = np.squeeze(res)
                out[i][idx] = np.asarray(res)[index]
        return out

    if n_output > 1:
        ufunc = _multi_ufunc
    else:
        ufunc = _ufunc

    update_docstring(ufunc, func, n_output)
    return ufunc


def update_docstring(ufunc, func, n_output=1):
    """Update ArviZ generated ufunc docstring."""
    module = ""
    name = ""
    docstring = ""
    if hasattr(func, "__module__") and isinstance(func.__module__, str):
        module += func.__module__
    if hasattr(func, "__name__"):
        name += func.__name__
    if hasattr(func, "__doc__") and isinstance(func.__doc__, str):
        docstring += func.__doc__
    ufunc.__doc__ += "\n\n"
    if module or name:
        ufunc.__doc__ += "This function is a ufunc wrapper for "
        ufunc.__doc__ += module + "." + name
        ufunc.__doc__ += "\n"
    ufunc.__doc__ += 'Call ufunc with n_args from xarray against "chain" and "draw" dimensions:'
    ufunc.__doc__ += "\n\n"
    input_core_dims = 'tuple(("chain", "draw") for _ in range(n_args))'
    if n_output > 1:
        output_core_dims = f" tuple([] for _ in range({n_output}))"
        msg = f"xr.apply_ufunc(ufunc, dataset, input_core_dims={input_core_dims}, "
        msg += f"output_core_dims={ output_core_dims})"
    else:
        output_core_dims = ""
        msg = f"xr.apply_ufunc(ufunc, dataset, input_core_dims={input_core_dims})"
    ufunc.__doc__ += msg
    ufunc.__doc__ += "\n\n"
    ufunc.__doc__ += "For example: np.std(data, ddof=1) --> n_args=2"
    if docstring:
        ufunc.__doc__ += "\n\n"
        ufunc.__doc__ += module
        ufunc.__doc__ += name
        ufunc.__doc__ += " docstring:"
        ufunc.__doc__ += "\n\n"
        ufunc.__doc__ += docstring


def logsumexp(ary, *, b=None, b_inv=None, axis=None, keepdims=False, out=None, copy=True):
    """Stable logsumexp when b >= 0 and b is scalar.

    b_inv overwrites b unless b_inv is None.
    """
    # check dimensions for result arrays
    ary = np.asarray(ary)
    if ary.dtype.kind == "i":
        ary = ary.astype(np.float64)
    dtype = ary.dtype.type
    shape = ary.shape
    shape_len = len(shape)
    if isinstance(axis, Sequence):
        axis = tuple(axis_i if axis_i >= 0 else shape_len + axis_i for axis_i in axis)
        agroup = axis
    else:
        axis = axis if (axis is None) or (axis >= 0) else shape_len + axis
        agroup = (axis,)
    shape_max = (
        tuple(1 for _ in shape)
        if axis is None
        else tuple(1 if i in agroup else d for i, d in enumerate(shape))
    )
    # create result arrays
    if out is None:
        if not keepdims:
            out_shape = (
                tuple()
                if axis is None
                else tuple(d for i, d in enumerate(shape) if i not in agroup)
            )
        else:
            out_shape = shape_max
        out = np.empty(out_shape, dtype=dtype)
    if b_inv == 0:
        return np.full_like(out, np.inf, dtype=dtype) if out.shape else np.inf
    if b_inv is None and b == 0:
        return np.full_like(out, -np.inf) if out.shape else -np.inf
    ary_max = np.empty(shape_max, dtype=dtype)
    # calculations
    ary.max(axis=axis, keepdims=True, out=ary_max)
    if copy:
        ary = ary.copy()
    ary -= ary_max
    np.exp(ary, out=ary)
    ary.sum(axis=axis, keepdims=keepdims, out=out)
    np.log(out, out=out)
    if b_inv is not None:
        ary_max -= np.log(b_inv)
    elif b:
        ary_max += np.log(b)
    out += ary_max if keepdims else ary_max.squeeze()
    # transform to scalar if possible
    return out if out.shape else dtype(out)


def not_valid(ary, check_nan=True, check_shape=True, nan_kwargs=None, shape_kwargs=None):
    """Validate ndarray.

    Parameters
    ----------
    ary : numpy.ndarray
    check_nan : bool
        Check if any value contains NaN.
    check_shape : bool
        Check if array has correct shape. Assumes dimensions in order (chain, draw, *shape).
        For 1D arrays (shape = (n,)) assumes chain equals 1.
    nan_kwargs : dict
        Valid kwargs are:
            axis : int,
                Defaults to None.
            how : str, {"all", "any"}
                Default to "any".
    shape_kwargs : dict
        Valid kwargs are:
            min_chains : int
                Defaults to 1.
            min_draws : int
                Defaults to 4.

    Returns
    -------
    bool
    """
    ary = np.asarray(ary)

    nan_error = False
    draw_error = False
    chain_error = False

    # for arviz-plots alignment, if all elements are nan return nan without indicating
    # any error
    isnan = np.isnan(ary)
    if isnan.all():
        return True

    if check_nan:
        if nan_kwargs is None:
            nan_kwargs = {}

        axis = nan_kwargs.get("axis", None)
        if nan_kwargs.get("how", "any").lower() == "all":
            nan_error = isnan.all(axis)
        else:
            nan_error = isnan.any(axis)

        if (isinstance(nan_error, bool) and nan_error) or nan_error.any():
            _log.warning("Array contains NaN-value.")

    if check_shape:
        shape = ary.shape

        if shape_kwargs is None:
            shape_kwargs = {}

        min_chains = shape_kwargs.get("min_chains", 2)
        min_draws = shape_kwargs.get("min_draws", 4)
        error_msg = f"Shape validation failed: input_shape: {shape}, "
        error_msg += f"minimum_shape: (chains={min_chains}, draws={min_draws})"

        chain_error = ((min_chains > 1) and (len(shape) < 2)) or (shape[0] < min_chains)
        draw_error = ((len(shape) < 2) and (shape[0] < min_draws)) or (
            (len(shape) > 1) and (shape[1] < min_draws)
        )

        if chain_error or draw_error:
            _log.warning(error_msg)

    return nan_error | chain_error | draw_error


def smooth_data(obs_vals, pp_vals):
    """Smooth data, helper function for discrete data in plot_pbv, loo_pit and plot_loo_pit."""
    x = np.linspace(0, 1, len(obs_vals))
    csi = CubicSpline(x, obs_vals)
    obs_vals = csi(np.linspace(0.01, 0.99, len(obs_vals)))

    x = np.linspace(0, 1, pp_vals.shape[1])
    csi = CubicSpline(x, pp_vals, axis=1)
    pp_vals = csi(np.linspace(0.01, 0.99, pp_vals.shape[1]))

    return obs_vals, pp_vals



"""Class with array functions.

"array" functions work on any dimension array,
batching as necessary.
"""

import numpy as np

from arviz_stats.base.density import _DensityBase
from arviz_stats.base.diagnostics import _DiagnosticsBase
from arviz_stats.base.stats_utils import make_ufunc


def process_ary_axes(ary, axes):
    """Process input array and axes to ensure input core dims are the last ones.

    Parameters
    ----------
    ary : array_like
    axes : int or sequence of int
    """
    if axes is None:
        axes = list(range(ary.ndim))
    if isinstance(axes, int):
        axes = [axes]
    axes = [ax if ax >= 0 else ary.ndim + ax for ax in axes]
    reordered_axes = [i for i in range(ary.ndim) if i not in axes] + list(axes)
    ary = np.transpose(ary, axes=reordered_axes)
    return ary, np.arange(-len(axes), 0, dtype=int)


class BaseArray(_DensityBase, _DiagnosticsBase):
    """Class with numpy+scipy only functions that take array inputs.

    Notes
    -----
    If a new dimension is created by the function it must be added at the end of the array.
    Otherwise the functions won't be compatible with :func:`xarray.apply_ufunc`.
    """

    def hdi(
        self,
        ary,
        prob,
        axes=-1,
        method="nearest",
        circular=False,
        max_modes=10,
        skipna=False,
        **kwargs,
    ):
        """Compute HDI function on array-like input."""
        if not 1 >= prob > 0:
            raise ValueError("The value of `prob` must be in the (0, 1] interval.")
        ary, axes = process_ary_axes(ary, axes)
        is_discrete = np.issubdtype(ary.dtype, np.integer) or np.issubdtype(ary.dtype, np.bool_)
        is_multimodal = method.startswith("multimodal")
        if is_multimodal and circular and is_discrete:
            raise ValueError("Multimodal hdi not supported for discrete circular data.")
        hdi_func = {
            "nearest": self._hdi_nearest,
            "multimodal": (
                self._hdi_multimodal_discrete if is_discrete else self._hdi_multimodal_continuous
            ),
            "multimodal_sample": (
                self._hdi_multimodal_discrete if is_discrete else self._hdi_multimodal_continuous
            ),
        }[method]
        hdi_array = make_ufunc(
            hdi_func,
            n_output=1,
            n_input=1,
            n_dims=len(axes),
        )
        func_kwargs = {
            "prob": prob,
            "skipna": skipna,
            "out_shape": (max_modes, 2) if is_multimodal else (2,),
            "circular": circular,
        }
        if is_multimodal:
            func_kwargs["max_modes"] = max_modes
            if is_discrete:
                func_kwargs.pop("circular")
                func_kwargs.pop("skipna")
            else:
                func_kwargs["bw"] = "isj" if not circular else "taylor"
            func_kwargs.update(kwargs)

        if method == "multimodal_sample":
            func_kwargs["from_sample"] = True

        result = hdi_array(ary, **func_kwargs)
        if is_multimodal:
            mode_mask = [np.all(np.isnan(result[..., i, :])) for i in range(result.shape[-2])]
            result = result[..., ~np.array(mode_mask), :]
        return result

    def ess(self, ary, chain_axis=-2, draw_axis=-1, method="bulk", relative=False, prob=None):
        """Compute of ess on array-like inputs."""
        method = method.lower()
        # fmt: off
        valid_methods = {
            "bulk", "tail", "mean", "sd", "quantile", "local", "median", "mad",
            "z_scale", "folded", "identity"
        }
        # fmt: on
        if method not in valid_methods:
            raise ValueError(f"Requested method '{method}' but it must be one of {valid_methods}")
        if chain_axis is None:
            ary = np.expand_dims(ary, axis=0)
            chain_axis = 0
        ary, _ = process_ary_axes(ary, [chain_axis, draw_axis])
        ess_func = getattr(self, f"_ess_{method}")
        func_kwargs = {"relative": relative}
        if prob is not None:
            func_kwargs["prob"] = prob
        ess_array = make_ufunc(ess_func, n_output=1, n_input=1, n_dims=2, ravel=False)
        return ess_array(ary, **func_kwargs)

    def rhat(self, ary, chain_axis=-2, draw_axis=-1, method="rank"):
        """Compute of rhat on array-like inputs."""
        method = method.lower()
        valid_methods = {"rank", "folded", "z_scale", "split", "identity"}
        if method not in valid_methods:
            raise ValueError(f"Requested method '{method}' but it must be one of {valid_methods}")
        if chain_axis is None:
            ary = np.expand_dims(ary, axis=0)
            chain_axis = 0
        ary, _ = process_ary_axes(ary, [chain_axis, draw_axis])
        rhat_func = getattr(self, f"_rhat_{method}")
        rhat_array = make_ufunc(rhat_func, n_output=1, n_input=1, n_dims=2, ravel=False)
        return rhat_array(ary)

    def mcse(self, ary, chain_axis=-2, draw_axis=-1, method="mean", prob=None):
        """Compute of mcse on array-like inputs."""
        method = method.lower()
        valid_methods = {"mean", "sd", "median", "quantile"}
        if method not in valid_methods:
            raise ValueError(f"Requested method '{method}' but it must be one of {valid_methods}")
        if chain_axis is None:
            ary = np.expand_dims(ary, axis=0)
            chain_axis = 0
        ary, _ = process_ary_axes(ary, [chain_axis, draw_axis])
        mcse_func = getattr(self, f"_mcse_{method}")
        func_kwargs = {} if prob is None else {"prob": prob}
        mcse_array = make_ufunc(mcse_func, n_output=1, n_input=1, n_dims=2, ravel=False)
        return mcse_array(ary, **func_kwargs)

    def pareto_min_ss(self, ary, chain_axis=-2, draw_axis=-1):
        """Compute minimum effective sample size."""
        if chain_axis is None:
            ary = np.expand_dims(ary, axis=0)
            chain_axis = 0
        ary, _ = process_ary_axes(ary, [chain_axis, draw_axis])
        pms_array = make_ufunc(self._pareto_min_ss, n_output=1, n_input=1, n_dims=2, ravel=False)
        return pms_array(ary)

    def psislw(self, ary, r_eff=1, axes=-1):
        """-"""
        ary, axes = process_ary_axes(ary, axes)
        psl_ufunc = make_ufunc(
            self._psislw,
            n_output=2,
            n_input=1,
            n_dims=len(axes),
            ravel=False,
        )
        return psl_ufunc(ary, out_shape=[(ary.shape[i] for i in axes), []], r_eff=r_eff)

    def power_scale_lw(self, ary, alpha=0, axes=-1):
        """Compute log weights for power-scaling component by alpha."""
        ary, axes = process_ary_axes(ary, axes)
        psl_ufunc = make_ufunc(
            self._power_scale_lw,
            n_output=1,
            n_input=1,
            n_dims=len(axes),
            ravel=False,
        )
        return psl_ufunc(ary, out_shape=(ary.shape[i] for i in axes), alpha=alpha)

    def power_scale_sense(
        self, ary, lower_w, upper_w, lower_alpha, upper_alpha, chain_axis=-2, draw_axis=-1
    ):
        """Compute power-scaling sensitivity."""
        if chain_axis is None:
            ary = np.expand_dims(ary, axis=0)
            lower_w = np.expand_dims(lower_w, axis=0)
            upper_w = np.expand_dims(upper_w, axis=0)
            chain_axis = 0
        ary, _ = process_ary_axes(ary, [chain_axis, draw_axis])
        lower_w, _ = process_ary_axes(lower_w, [chain_axis, draw_axis])
        upper_w, _ = process_ary_axes(upper_w, [chain_axis, draw_axis])
        pss_array = make_ufunc(
            self._power_scale_sense, n_output=1, n_input=3, n_dims=2, ravel=False
        )
        return pss_array(ary, lower_w, upper_w, lower_alpha=lower_alpha, upper_alpha=upper_alpha)

    def compute_ranks(self, ary, axes=-1, relative=False):
        """Compute ranks of MCMC samples."""
        ary, axes = process_ary_axes(ary, axes)
        compute_ranks_ufunc = make_ufunc(
            self._compute_ranks,
            n_output=1,
            n_input=1,
            n_dims=len(axes),
            ravel=False,
        )
        return compute_ranks_ufunc(ary, out_shape=(ary.shape[i] for i in axes), relative=relative)

    def get_bins(self, ary, axes=-1, bins="arviz"):
        """Compute default bins."""
        ary, axes = process_ary_axes(ary, axes)
        get_bininfo_ufunc = make_ufunc(
            self._get_bininfo,
            n_output=3,
            n_input=1,
            n_dims=len(axes),
        )
        # TODO: improve handling of array_like bins
        x_min, x_max, width = get_bininfo_ufunc(ary, bins=bins)
        n_bins = np.ceil((x_max - x_min) / width)
        n_bins = np.ceil(np.mean(n_bins)).astype(int)
        return np.moveaxis(np.linspace(x_min, x_max, n_bins + 1), 0, -1)

    # pylint: disable=redefined-builtin, too-many-return-statements
    # noqa: PLR0911
    def histogram(self, ary, bins=None, range=None, weights=None, axes=-1, density=None):
        """Compute histogram over provided axes."""
        if isinstance(axes, int):
            axes = [axes]
        axes = [ax if ax >= 0 else ary.ndim + ax for ax in axes]
        reordered_axes = [i for i in np.arange(ary.ndim) if i not in axes] + list(axes)
        if weights is not None:
            assert ary.shape == weights.shape
            weights = np.transpose(weights, axes=reordered_axes)
        ary = np.transpose(ary, axes=reordered_axes)
        broadcased_shape = ary.shape[: -len(axes)]
        if bins is None:
            bins = self.get_bins(ary, axes=np.arange(-len(axes), 0, dtype=int))
        if isinstance(bins, int | str):
            # avoid broadcasting over bins -> can't be positional argument
            if (range is None) or (np.size(range) == 2):
                # avoid broadcasting over range
                if weights is not None:
                    # ensure broadcasting over weights
                    histogram_ufunc = make_ufunc(
                        lambda ary, weights: self._histogram(
                            ary, bins=bins, range=range, weights=weights, density=density
                        ),
                        n_output=2,
                        n_input=2,
                        n_dims=len(axes),
                    )
                    return histogram_ufunc(ary, weights, shape_from_1st=True)
                # avoid broadcasting over weights -> no broadcasting anywhere
                histogram_ufunc = make_ufunc(
                    self._histogram, n_output=2, n_input=1, n_dims=len(axes)
                )
                return histogram_ufunc(
                    ary, bins=bins, range=range, density=density, shape_from_1st=True
                )
            # ensure broadcasting over range
            assert range.shape[:-1] == broadcased_shape
            if weights is not None:
                # ensure broadcasting over weights
                histogram_ufunc = make_ufunc(
                    lambda ary, range, weights: self._histogram(
                        ary, bins=bins, range=range, weights=weights, density=density
                    ),
                    n_output=2,
                    n_input=3,
                    n_dims=len(axes),
                )
                return histogram_ufunc(ary, range, weights, shape_from_1st=True)
            # avoid broadcasting over weights while broadcasting over range
            histogram_ufunc = make_ufunc(
                lambda ary, range: self._histogram(ary, bins=bins, range=range, density=density),
                n_output=2,
                n_input=2,
                n_dims=len(axes),
            )
            return histogram_ufunc(ary, range, shape_from_1st=True)
        # ensure broadcasting over bins
        assert bins.shape[:-1] == broadcased_shape
        if (range is None) or (np.size(range) == 2):
            # avoid broadcasting over range
            if weights is not None:
                # ensure broadcasting over weights
                histogram_ufunc = make_ufunc(
                    lambda ary, bins, weights: self._histogram(
                        ary, bins=bins, range=range, weights=weights, density=density
                    ),
                    n_output=2,
                    n_input=3,
                    n_dims=len(axes),
                )
                return histogram_ufunc(ary, bins, weights, shape_from_1st=True)
            # avoid broadcasting over weights
            histogram_ufunc = make_ufunc(
                lambda ary, bins: self._histogram(ary, bins=bins, range=range, density=density),
                n_output=2,
                n_input=2,
                n_dims=len(axes),
            )
            return histogram_ufunc(ary, bins, shape_from_1st=True)
        # ensure broadcasting over range
        assert range.shape[:-1] == broadcased_shape
        if weights is not None:
            # ensure broadcasting over weights
            histogram_ufunc = make_ufunc(
                lambda ary, bins, range, weights: self._histogram(
                    ary, bins=bins, range=range, weights=weights, density=density
                ),
                n_output=2,
                n_input=4,
                n_dims=len(axes),
            )
            return histogram_ufunc(ary, bins, range, weights, shape_from_1st=True)
        # avoid broadcasting over weights
        histogram_ufunc = make_ufunc(
            lambda ary, bins, range: self._histogram(ary, bins=bins, range=range, density=density),
            n_output=2,
            n_input=3,
            n_dims=len(axes),
        )
        return histogram_ufunc(ary, bins, range, shape_from_1st=True)

    def kde(self, ary, axes=-1, circular=False, grid_len=512, **kwargs):
        """Compute of kde on array-like inputs."""
        ary, axes = process_ary_axes(ary, axes)
        kde_ufunc = make_ufunc(
            self._kde,
            n_output=3,
            n_input=1,
            n_dims=len(axes),
        )
        return kde_ufunc(
            ary,
            out_shape=((grid_len,), (grid_len,), ()),
            grid_len=grid_len,
            circular=circular,
            **kwargs,
        )


array_stats = BaseArray()



"""Utility functions for computing the PIT ECDF and the simultaneous confidence bands."""

import math

import numpy as np
from arviz_base import dict_to_dataset
from numba import vectorize
from scipy.optimize import minimize_scalar
from scipy.special import bdtr, bdtrik  # pylint: disable=no-name-in-module
from scipy.stats import uniform


def difference_ecdf_pit(dt, data_pairs, ci_prob, coverage, randomized, method, n_simulations):
    """Compute the difference PIT ECDF values.

    The probability of the posterior predictive being less than or equal to the observed data.
    should be uniformly distributed. This function computes the PIT ECDF and the simultaneous
    confidence bands.

    Parameters
    ----------
    dt: DataTree
        DataTree with "posterior_predictive" and "observed_data" groups
    data_pairs : tuple
        Tuple with first element contains the posterior predictive name (or list of names)
        and the second element contains the observed data variable name (or list of names).
    ci_prob : float, optional
        The probability for the credible interval.
    randomized : list of bool
        Whether to randomize the PIT values. Randomization is needed for discrete data.
    n_simulations : int
        The number of simulations to use with method `simulation`.
    method : str
        The method to use to compute the confidence bands. Either `simulation` or `optimized`.
    """
    dictio = {}
    rng = np.random.default_rng(214)

    for idx, (var_predictive, var_obs) in enumerate(data_pairs.items()):
        vals = (dt.posterior_predictive[var_predictive] <= dt.observed_data[var_obs]).mean(
            ("chain", "draw")
        )
        if randomized[idx]:
            vals_less = (dt.posterior_predictive[var_predictive] < dt.observed_data[var_obs]).mean(
                ("chain", "draw")
            )

            urvs = rng.uniform(size=vals.shape)
            vals = urvs * vals_less + (1 - urvs) * vals

        if coverage:
            vals = 2 * np.abs(vals - 0.5)

        eval_points, ecdf, ci_lb, ci_ub = ecdf_pit(vals, ci_prob, method, n_simulations, rng)
        dictio[var_predictive] = np.stack(
            [eval_points, ecdf - eval_points, ci_lb - eval_points, ci_ub - eval_points]
        )

    return (
        dict_to_dataset(dictio)
        .rename({"draw": "x_values", "chain": "plot_axis"})
        .assign_coords({"plot_axis": ["x", "y", "y_bottom", "y_top"]})
    )


def ecdf_pit(vals, ci_prob, method, n_simulations, rng=None):
    """
    Compute the PIT ECDF and the pointwise confidence bands.

    Parameters
    ----------
    vals : array-like
        The values to compute the PIT ECDF.
    ci_prob : float, optional
        The probability for the credible interval.
    method : str
        The method to use to compute the confidence bands. Either `simulation` or `optimized`.
    n_simulations : int
        The number of simulations to use with method `simulation`.
    rng : Generator, optional
        The random number generator to use with the simulation method.

    Returns
    -------
    eval_points : array-like
        The evaluation points.
    ecdf : array-like
        The PIT ECDF.
    lower : array-like
        The lower confidence band.
    upper : array-like
        The upper confidence band.
    """
    if rng is None:
        rng = np.random.default_rng(214)

    n_draws = len(vals)
    eval_points = np.linspace(1 / n_draws, 1, n_draws)
    ecdf = compute_ecdf(vals, eval_points)

    cdf_at_eval_points = uniform(0, 1).cdf(eval_points)

    if method == "simulation":
        prob_pointwise = simulate_confidence_bands(
            n_draws,
            eval_points,
            cdf_at_eval_points,
            ci_prob,
            n_simulations,
            rng,
        )

    elif method == "optimized":
        prob_pointwise = optimize_confidence_bands(n_draws, cdf_at_eval_points, ci_prob)
    else:
        raise ValueError(f"Invalid method: {method}")

    lower, upper = get_pointwise_confidence_band(prob_pointwise, n_draws, cdf_at_eval_points)

    return eval_points, ecdf, lower, upper


def compute_ecdf(sample, eval_points):
    """Compute the empirical CDF at the evaluation points."""
    sample = np.sort(sample)
    return np.searchsorted(sample, eval_points, side="right") / len(sample)


def get_pointwise_confidence_band(prob, ndraws, cdf_at_eval_points):
    """Compute the `prob`-level pointwise confidence band."""
    lower_ci = (1 - prob) / 2
    upper_ci = 1 - lower_ci

    # We use the bdtrik function instead of instantiating a binomial distribution
    # and computing the quantiles because it is faster
    count_lower = np.ceil(bdtrik(lower_ci, ndraws, cdf_at_eval_points))
    count_upper = np.ceil(bdtrik(upper_ci, ndraws, cdf_at_eval_points))

    prob_lower = count_lower / ndraws
    prob_upper = count_upper / ndraws
    return prob_lower, prob_upper


### Simulation method
def simulate_confidence_bands(n_draws, eval_points, cdf_at_eval_points, prob, n_simulations, rng):
    """Simulate method for simultaneous confidence bands.

    Compute the smallest marginal probability of a pointwise confidence band that
    contains the ECDF using the simulation method.

    Parameters
    ----------
    n_draws : int
        The number of draws.
    eval_points : array-like
        The evaluation points.
    cdf_at_eval_points : array-like
        The CDF at the evaluation points.
    prob : float
        The probability for the credible interval.
    num_trials : int
        The number of trials to use.
    rng : Generator
        The random number generator to use.
    """
    probs_pointwise = np.empty(n_simulations)
    for i in range(n_simulations):
        sample = rng.uniform(0, 1, size=n_draws)
        ecdf_at_eval_points = compute_ecdf(sample, eval_points)
        prob_pointwise = fit_pointwise_band_probability(
            n_draws, ecdf_at_eval_points, cdf_at_eval_points
        )
        probs_pointwise[i] = prob_pointwise

    return np.quantile(probs_pointwise, prob)


def fit_pointwise_band_probability(
    n_draws,
    ecdf_at_eval_points,
    cdf_at_eval_points,
):
    """Return the smallest marginal prob of confidence band that contains the ECDF.

    Parameters
    ----------
    n_draws : int
        The number of draws.
    ecdf_at_eval_points : array-like
        The ECDF at the evaluation points.
    cdf_at_eval_points : array-like
        The CDF at the evaluation points.
    """
    ecdf_scaled = n_draws * ecdf_at_eval_points
    # We use the bdtr function instead of instantiating a binomial distribution
    # and computing the cdf because it is faster
    prob_lower_tail = np.min(bdtr(ecdf_scaled, n_draws, cdf_at_eval_points))
    prob_upper_tail = np.min(1 - bdtr(ecdf_scaled - 1, n_draws, cdf_at_eval_points))
    prob_pointwise = 1 - 2 * min(prob_lower_tail, prob_upper_tail)
    return prob_pointwise


### Optimization method
### The optimization method should be faster than the simulation method
### but it is not the case with the current implementation
### Leave it here for future reference
def optimize_confidence_bands(ndraws, cdf_at_eval_points, prob):
    """Estimate probability for simultaneous confidence band using optimization.

    This function simulates the pointwise probability needed to construct pointwise confidence bands
    that form a `prob`-level confidence envelope for the ECDF of a sample.
    """
    # cdf_at_eval_points = np.unique(cdf_at_eval_points)
    prob_pointwise = minimize_scalar(
        ecdf_band_optimization_objective,
        bounds=(prob, 1),
        method="bounded",
        options={"xatol": 0.001},
        args=(cdf_at_eval_points, ndraws, prob),
    ).x
    return prob_pointwise


@vectorize()
def binomial_pmf(k, n, p, loc):
    """Compute the binomial probability mass function."""
    k -= loc
    if k < 0 or k > n:
        return 0.0
    if p == 0:
        return 1.0 if k == 0 else 0.0
    if p == 1:
        return 1.0 if k == n else 0.0
    if k == 0:
        return (1 - p) ** n
    if k == n:
        return p**n
    lbinom = math.lgamma(n + 1) - math.lgamma(k + 1) - math.lgamma(n - k + 1)
    return np.exp(lbinom + k * np.log(p) + (n - k) * np.log1p(-p))


def ecdf_band_optimization_objective(prob_pointwise, cdf_at_eval_points, ndraws, prob_target):
    """Objective function for optimizing the simultaneous confidence band probability."""
    lower, upper = get_pointwise_confidence_band(prob_pointwise, ndraws, cdf_at_eval_points)
    lower_count = (lower * ndraws).astype(int)
    upper_count = (upper * ndraws).astype(int) + 1
    cdf_with_zero = np.insert(cdf_at_eval_points[:-1], 0, 0)
    prob_between_points = (cdf_at_eval_points - cdf_with_zero) / (1 - cdf_with_zero)
    prob_interior = ecdf_band_interior_probability(
        prob_between_points, ndraws, lower_count, upper_count
    )
    return abs(prob_interior - prob_target)


def ecdf_band_interior_probability(prob_between_points, ndraws, lower_count, upper_count):
    """Compute the probability of the ECDF being inside the pointwise confidence bands."""
    interval_left = np.zeros(1)
    prob_interior = np.ones(1)
    for i in range(prob_between_points.shape[0]):
        interval_right = np.arange(lower_count[i], upper_count[i])
        prob_interior = update_ecdf_band_interior_probabilities(
            prob_interior, interval_left, interval_right, prob_between_points[i], ndraws
        )
        interval_left = interval_right
    return prob_interior.sum()


def update_ecdf_band_interior_probabilities(prob_left, interval_left, interval_right, p, ndraws):
    """Update the probability of the ECDF being inside the pointwise confidence bands."""
    interval_left = interval_left[:, None]
    prob_conditional = binomial_pmf(interval_right, ndraws - interval_left, p, interval_left)
    prob_right = np.dot(prob_left, prob_conditional)
    return prob_right



# pylint: disable=invalid-name,too-many-lines
"""Density estimation functions for ArviZ."""

import math
import warnings

import numpy as np
from scipy.optimize import brentq
from scipy.signal import convolve, convolve2d
from scipy.signal.windows import gaussian
from scipy.sparse import coo_matrix
from scipy.special import ive  # pylint: disable=no-name-in-module

from arviz_stats.base.core import _CoreBase


class _DensityBase(_CoreBase):
    """Class with numpy+scipy only density related functions."""

    bw_methods_linear = ("scott", "silverman", "isj", "experimental")

    def dct1d(self, x):
        """Discrete Cosine Transform in 1 Dimension.

        Parameters
        ----------
        x : numpy array
            1 dimensional array of values for which the
            DCT is desired

        Returns
        -------
        output : DTC transformed values
        """
        x_len = len(x)

        even_increasing = np.arange(0, x_len, 2)
        odd_decreasing = np.arange(x_len - 1, 0, -2)

        x = np.concatenate((x[even_increasing], x[odd_decreasing]))

        w_1k = np.r_[1, (2 * np.exp(-(0 + 1j) * (np.arange(1, x_len)) * np.pi / (2 * x_len)))]
        output = np.real(w_1k * self.fft(x))

        return output

    def _fixed_point(self, t, n, k_sq, a_sq):  # pylint: disable=no-self-use
        """Calculate t-zeta*gamma^[l](t).

        Implementation of the function t-zeta*gamma^[l](t) derived from equation (30) in [1].

        References
        ----------
        .. [1] Kernel density estimation via diffusion.
        Z. I. Botev, J. F. Grotowski, and D. P. Kroese.
        Ann. Statist. 38 (2010), no. 5, 2916--2957.
        """
        k_sq = np.asarray(k_sq, dtype=np.float64)
        a_sq = np.asarray(a_sq, dtype=np.float64)

        l = 7  # noqa: E741
        f = np.sum(np.power(k_sq, l) * a_sq * np.exp(-k_sq * np.pi**2 * t))
        f *= 0.5 * np.pi ** (2.0 * l)

        for j in np.arange(l - 1, 2 - 1, -1):
            c1 = (1 + 0.5 ** (j + 0.5)) / 3
            c2 = np.prod(np.arange(1.0, 2 * j + 1, 2, dtype=np.float64))
            c2 /= (np.pi / 2) ** 0.5
            t_j = np.power((c1 * (c2 / (n * f))), (2.0 / (3.0 + 2.0 * j)))
            f = np.sum(k_sq**j * a_sq * np.exp(-k_sq * np.pi**2.0 * t_j))
            f *= 0.5 * np.pi ** (2 * j)

        out = t - (2 * n * np.pi**0.5 * f) ** (-0.4)
        return out

    def _root(self, function, n, args, x, x_range):
        # The right bound is at most 0.01
        found = False
        n = max(min(1050, n), 50)
        tol = 10e-12 + 0.01 * (n - 50) / 1000

        while not found:
            try:
                bw, res = brentq(function, 0, 0.01, args=args, full_output=True, disp=False)
                found = res.converged
            except ValueError:
                bw = 0
                tol *= 2.0
                found = False
            if bw <= 0 or tol >= 1:
                bw = (self.bw_silverman(x) / x_range) ** 2
                return bw
        return bw

    def bw_scott(self, x, x_std=None, **kwargs):  # pylint: disable=unused-argument, no-self-use
        """Scott's Rule."""
        if x_std is None:
            x_std = np.std(x)
        bw = 1.06 * x_std * len(x) ** (-0.2)
        return bw

    def bw_silverman(self, x, x_std=None, **kwargs):  # pylint: disable=unused-argument
        """Silverman's Rule."""
        if x_std is None:
            x_std = np.std(x)
        q25, q75 = self.quantile(x, [0.25, 0.75])
        x_iqr = q75 - q25
        a = min(x_std, x_iqr / 1.34)
        bw = 0.9 * a * len(x) ** (-0.2)
        return bw

    def bw_isj(self, x, grid_counts=None, x_std=None, x_range=None):
        """Improved Sheather-Jones bandwidth estimation.

        Notes
        -----
        Improved Sheather and Jones method as explained in [1]_.
        This method is used internally by the KDE estimator, resulting in saved computation time
        as minimums, maximums and the grid are pre-computed.

        References
        ----------
        .. [1] Kernel density estimation via diffusion.
        Z. I. Botev, J. F. Grotowski, and D. P. Kroese.
        Ann. Statist. 38 (2010), no. 5, 2916--2957.
        """
        x_len = len(x)
        if x_range is None:
            x_min = np.min(x)
            x_max = np.max(x)
            x_range = x_max - x_min

        # Relative frequency per bin
        if grid_counts is None:
            x_std = np.std(x)
            grid_len = 256
            grid_min = x_min - 0.5 * x_std
            grid_max = x_max + 0.5 * x_std
            grid_counts, _ = self._histogram(
                x, bins=grid_len, range=(grid_min, grid_max), density=False
            )
        else:
            grid_len = len(grid_counts) - 1

        grid_relfreq = grid_counts / x_len

        # Discrete cosine transform of the data
        a_k = self.dct1d(grid_relfreq)

        k_sq = np.arange(1, grid_len) ** 2
        a_sq = a_k[range(1, grid_len)] ** 2

        t = self._root(self._fixed_point, x_len, args=(x_len, k_sq, a_sq), x=x, x_range=x_range)
        h = t**0.5 * x_range
        return h

    def bw_experimental(self, x, grid_counts=None, x_std=None, x_range=None):
        """Experimental bandwidth estimator."""
        bw_silverman = self.bw_silverman(x, x_std=x_std)
        bw_isj = self.bw_isj(x, grid_counts=grid_counts, x_range=x_range)
        return 0.5 * (bw_silverman + bw_isj)

    def get_bw(self, x, bw, grid_counts=None, x_std=None, x_range=None):
        """Compute bandwidth for a given data `x` and `bw`.

        Also checks `bw` is correctly specified.

        Parameters
        ----------
        x : 1-D numpy array
            1 dimensional array of sample data from the
            variable for which a density estimate is desired.
        bw: int, float or str
            If numeric, indicates the bandwidth and must be positive.
            If str, indicates the method to estimate the bandwidth.

        Returns
        -------
        bw: float
            Bandwidth
        """
        if isinstance(bw, bool):
            raise ValueError(
                "`bw` must not be of type `bool`.\n"
                "Expected a positive numeric or one of the following strings:\n"
                f"{self.bw_methods_linear}."
            )
        if isinstance(bw, int | float):
            if bw < 0:
                raise ValueError(f"Numeric `bw` must be positive.\nInput: {bw:.4f}.")
        elif isinstance(bw, str):
            bw_lower = bw.lower()

            if bw_lower not in self.bw_methods_linear:
                raise ValueError(
                    "Unrecognized bandwidth method.\n"
                    f"Input is: {bw_lower}.\n"
                    f"Expected one of: {self.bw_methods_linear}."
                )

            bw_fun = getattr(self, f"bw_{bw}")
            bw = bw_fun(x, grid_counts=grid_counts, x_std=x_std, x_range=x_range)
        else:
            raise ValueError(
                "Unrecognized `bw` argument.\n"
                "Expected a positive numeric or one of the following strings:\n"
                f"{self.bw_methods_linear}."
            )
        return bw

    def _normalize_angle(self, x, zero_centered=True):  # pylint: disable=no-self-use
        """Normalize angles.

        Normalize angles in radians to [-pi, pi) or [0, 2 * pi) according to `zero_centered`.
        """
        if zero_centered:
            return (x + np.pi) % (2 * np.pi) - np.pi
        return x % (2 * np.pi)

    def _vonmises_pdf(self, x, mu, kappa):  # pylint: disable=no-self-use
        """Calculate vonmises_pdf."""
        if kappa <= 0:
            raise ValueError("Argument 'kappa' must be positive.")
        pdf = 1 / (2 * np.pi * ive(0, kappa)) * np.exp(np.cos(x - mu) - 1) ** kappa
        return pdf

    def _a1inv(self, x):  # pylint: disable=no-self-use
        """Compute inverse function.

        Inverse function of the ratio of the first and
        zeroth order Bessel functions of the first kind.

        Returns the value k, such that a1inv(x) = k, i.e. a1(k) = x.
        """
        if 0 <= x < 0.53:
            return 2 * x + x**3 + (5 * x**5) / 6
        if x < 0.85:
            return -0.4 + 1.39 * x + 0.43 / (1 - x)
        return 1 / (x**3 - 4 * x**2 + 3 * x)

    def _kappa_mle(self, x):
        mean = self.circular_mean(x)
        kappa = self._a1inv(np.mean(np.cos(x - mean)))
        return kappa

    def bw_taylor(self, x):
        """Taylor's rule for circular bandwidth estimation.

        Notes
        -----
        This function implements a rule-of-thumb for choosing the bandwidth of a von Mises kernel
        density estimator that assumes the underlying distribution is von Mises as introduced in
        Taylor (2008) [1]_.
        It is analogous to Scott's rule for the Gaussian KDE.

        Circular bandwidth has a different scale from linear bandwidth. Unlike linear scale, low
        bandwidths are associated with oversmoothing and high values with undersmoothing.

        References
        ----------
        .. [1] C.C Taylor (2008). Automatic bandwidth selection for circular
            density estimation.
            Computational Statistics and Data Analysis, 52, 7, 3493–3500.
        """
        x_len = len(x)
        kappa = self._kappa_mle(x)
        num = 3 * x_len * kappa**2 * ive(2, 2 * kappa)
        den = 4 * np.pi**0.5 * ive(0, kappa) ** 2
        return (num / den) ** 0.4

    def check_custom_lims(self, custom_lims, x_min, x_max):  # pylint: disable=no-self-use
        """Check if `custom_lims` are of the correct type.

        It accepts numeric lists/tuples of length 2.

        Parameters
        ----------
        custom_lims : Object whose type is checked.

        Returns
        -------
        None: Object of type None
        """
        if not isinstance(custom_lims, list | tuple):
            raise TypeError(
                "`custom_lims` must be a numeric list or tuple of length 2.\n"
                f"Not an object of {type(custom_lims)}."
            )

        if len(custom_lims) != 2:
            raise AttributeError(f"`len(custom_lims)` must be 2, not {len(custom_lims)}.")

        any_bool = any(isinstance(i, bool) for i in custom_lims)
        if any_bool:
            raise TypeError("Elements of `custom_lims` must be numeric or None, not bool.")

        custom_lims = list(custom_lims)  # convert to a mutable object
        if custom_lims[0] is None:
            custom_lims[0] = x_min

        if custom_lims[1] is None:
            custom_lims[1] = x_max

        all_numeric = all(isinstance(i, int | float | np.integer | np.number) for i in custom_lims)
        if not all_numeric:
            raise TypeError(
                "Elements of `custom_lims` must be numeric or None.\nAt least one of them is not."
            )

        if not custom_lims[0] < custom_lims[1]:
            raise ValueError("`custom_lims[0]` must be smaller than `custom_lims[1]`.")

        if custom_lims[0] > x_min or custom_lims[1] < x_max:
            raise ValueError("Some observations are outside `custom_lims` boundaries.")

        return custom_lims

    def get_grid(
        self,
        x_min,
        x_max,
        x_std,
        extend_fct,
        grid_len,
        custom_lims,
        extend=True,
        bound_correction=False,
    ):
        """Compute the grid that bins the data used to estimate the density function.

        Parameters
        ----------
        x_min : float
            Minimum value of the data
        x_max: float
            Maximum value of the data.
        x_std: float
            Standard deviation of the data.
        extend_fct: bool
            Indicates the factor by which `x_std` is multiplied
            to extend the range of the data.
        grid_len: int
            Number of bins
        custom_lims: tuple or list
            Custom limits for the domain of the density estimation.
            Must be numeric of length 2. Overrides `extend`.
        extend: bool, optional
            Whether to extend the range of the data or not.
            Default is True.
        bound_correction: bool, optional
            Whether the density estimations performs boundary correction or not.
            This does not impacts directly in the output, but is used
            to override `extend`. Overrides `extend`.
            Default is False.

        Returns
        -------
        grid_len: int
            Number of bins
        grid_min: float
            Minimum value of the grid
        grid_max: float
            Maximum value of the grid
        """
        # Set up number of bins.
        grid_len = max(int(grid_len), 100)

        # Set up domain
        if custom_lims is not None:
            custom_lims = self.check_custom_lims(custom_lims, x_min, x_max)
            grid_min = custom_lims[0]
            grid_max = custom_lims[1]
        elif extend and not bound_correction:
            grid_extend = extend_fct * x_std
            grid_min = x_min - grid_extend
            grid_max = x_max + grid_extend
        else:
            grid_min = x_min
            grid_max = x_max
        return grid_min, grid_max, grid_len

    def kde_linear(
        self,
        x,
        bw="experimental",
        adaptive=False,
        extend=False,
        bound_correction=True,
        extend_fct=0,
        bw_fct=1,
        custom_lims=None,
        cumulative=False,
        grid_len=512,
        **kwargs,  # pylint: disable=unused-argument
    ):
        """One dimensional density estimation for linear data.

        Given an array of data points `x` it returns an estimate of
        the probability density function that generated the samples in `x`.

        Parameters
        ----------
        x : 1D numpy array
            Data used to calculate the density estimation.
        bw: int, float or str, optional
            If numeric, indicates the bandwidth and must be positive.
            If str, indicates the method to estimate the bandwidth and must be one of "scott",
            "silverman", "isj" or "experimental". Defaults to "experimental".
        adaptive: boolean, optional
            Indicates if the bandwidth is adaptive or not.
            It is the recommended approach when there are multiple modes with different spread.
            It is not compatible with convolution. Defaults to False.
        extend: boolean, default False
            Whether to extend the observed range for `x` in the estimation.
            It extends each bound by a multiple of the standard deviation of `x`
            given by `extend_fct`. Defaults to False.
        bound_correction: boolean, optional
            Whether to perform boundary correction on the bounds of `x` or not.
            Defaults to True.
        extend_fct: float, optional
            Number of standard deviations used to widen the lower and upper bounds of `x`.
            Defaults to 0.5.
        bw_fct: float, optional
            A value that multiplies `bw` which enables tuning smoothness by hand.
            Must be positive. Values below 1 decrease smoothness while values above 1 decrease it.
            Defaults to 1 (no modification).
        custom_lims: list or tuple, optional
            A list or tuple of length 2 indicating custom bounds for the range of `x`.
            Defaults to None which disables custom bounds.
        cumulative: bool, optional
            Whether return the PDF or the cumulative PDF. Defaults to False.
        grid_len: int, optional
            The number of intervals used to bin the data points i.e. the length of the grid used in
            the estimation. Defaults to 512.

        Returns
        -------
        grid : Gridded numpy array for the x values.
        pdf : Numpy array for the density estimates.
        bw: optional, the estimated bandwidth.
        """
        # Check `bw_fct` is numeric and positive
        if not isinstance(bw_fct, int | float | np.integer | np.floating):
            raise TypeError(f"`bw_fct` must be a positive number, not an object of {type(bw_fct)}.")

        if bw_fct <= 0:
            raise ValueError(f"`bw_fct` must be a positive number, not {bw_fct}.")

        # Preliminary calculations
        x_min = x.min()
        x_max = x.max()
        x_std = np.std(x)
        x_range = x_max - x_min

        # Determine grid
        grid_min, grid_max, grid_len = self.get_grid(
            x_min, x_max, x_std, extend_fct, grid_len, custom_lims, extend, bound_correction
        )
        grid_counts, grid_edges = self._histogram(
            x, bins=grid_len, range=(grid_min, grid_max), density=False
        )

        # Bandwidth estimation
        bw = bw_fct * self.get_bw(x, bw, grid_counts, x_std, x_range)

        # Density estimation
        if adaptive:
            grid, pdf = self.kde_adaptive(
                x, bw, grid_edges, grid_counts, grid_len, bound_correction
            )
        else:
            grid, pdf = self.kde_convolution(
                x, bw, grid_edges, grid_counts, grid_len, bound_correction
            )

        if cumulative:
            pdf = pdf.cumsum() / pdf.sum()
        else:
            # explicitly normalize to 1
            bin_width = grid_edges[1] - grid_edges[0]
            pdf /= pdf.sum() * bin_width

        return grid, pdf, bw

    # pylint: disable=no-self-use, unused-argument
    def kde_convolution(self, x, bw, grid_edges, grid_counts, grid_len, bound_correction, **kwargs):
        """Kernel density with convolution.

        One dimensional Gaussian kernel density estimation via convolution of the binned relative
        frequencies and a Gaussian filter. This is an internal function used by `kde()`.
        """
        # Calculate relative frequencies per bin
        bin_width = grid_edges[1] - grid_edges[0]
        f = grid_counts / bin_width / len(x)

        # Bandwidth must consider the bin width
        bw /= bin_width

        # See: https://stackoverflow.com/questions/2773606/gaussian-filter-in-matlab

        grid = (grid_edges[1:] + grid_edges[:-1]) / 2

        kernel_n = 2 * math.ceil(4 * bw) + 1

        kernel = gaussian(kernel_n, bw)

        if bound_correction:
            npad = int(grid_len / 5)
            f = np.concatenate([f[npad - 1 :: -1], f, f[grid_len : grid_len - npad - 1 : -1]])
            pdf = convolve(f, kernel, mode="same", method="direct")[npad : npad + grid_len]
        else:
            pdf = convolve(f, kernel, mode="same", method="direct")
        pdf /= bw * (2 * np.pi) ** 0.5

        return grid, pdf

    def kde_adaptive(self, x, bw, grid_edges, grid_counts, grid_len, bound_correction, **kwargs):
        """Compute Adaptive Kernel Density Estimation.

        Notes
        -----
        One dimensional adaptive Gaussian kernel density estimation.
        The implementation uses the binning technique.
        Since there is not an unique `bw`, the convolution is not possible.
        The alternative implemented in this function is known as Abramson's method.

        This is an internal function used by `kde()`.
        """
        # Pilot computations used for bandwidth adjustment
        pilot_grid, pilot_pdf = self.kde_convolution(
            x, bw, grid_edges, grid_counts, grid_len, bound_correction
        )

        # Adds to avoid np.log(0) and zero division
        pilot_pdf += 1e-9

        # Determine the modification factors
        pdf_interp = np.interp(x, pilot_grid, pilot_pdf)
        geom_mean = np.exp(np.mean(np.log(pdf_interp)))

        # Power of c = 0.5 -> Abramson's method
        adj_factor = (geom_mean / pilot_pdf) ** 0.5
        bw_adj = bw * adj_factor

        # Estimation of Gaussian KDE via binned method (convolution not possible)
        grid = pilot_grid

        if bound_correction:
            grid_npad = int(grid_len / 5)
            grid_width = grid_edges[1] - grid_edges[0]
            grid_pad = grid_npad * grid_width
            grid_padded = np.linspace(
                grid_edges[0] - grid_pad,
                grid_edges[grid_len - 1] + grid_pad,
                num=grid_len + 2 * grid_npad,
            )
            grid_counts = np.concatenate(
                [
                    grid_counts[grid_npad - 1 :: -1],
                    grid_counts,
                    grid_counts[grid_len : grid_len - grid_npad - 1 : -1],
                ]
            )
            bw_adj = np.concatenate(
                [
                    bw_adj[grid_npad - 1 :: -1],
                    bw_adj,
                    bw_adj[grid_len : grid_len - grid_npad - 1 : -1],
                ]
            )
            pdf_mat = (grid_padded - grid_padded[:, None]) / bw_adj[:, None]
            pdf_mat = np.exp(-0.5 * pdf_mat**2) * grid_counts[:, None]
            pdf_mat /= (2 * np.pi) ** 0.5 * bw_adj[:, None]
            pdf = np.sum(pdf_mat[:, grid_npad : grid_npad + grid_len], axis=0) / len(x)

        else:
            pdf_mat = (grid - grid[:, None]) / bw_adj[:, None]
            pdf_mat = np.exp(-0.5 * pdf_mat**2) * grid_counts[:, None]
            pdf_mat /= (2 * np.pi) ** 0.5 * bw_adj[:, None]
            pdf = np.sum(pdf_mat, axis=0) / len(x)

        return grid, pdf

    def kde_circular(
        self,
        x,
        bw="taylor",
        bw_fct=1,
        custom_lims=None,
        cumulative=False,
        grid_len=512,
        **kwargs,  # pylint: disable=unused-argument
    ):
        """One dimensional density estimation for circular data.

        Given an array of data points `x` measured in radians, it returns an estimate of the
        probability density function that generated the samples in `x`.

        Parameters
        ----------
        x : 1D numpy array
            Data used to calculate the density estimation.
        bw: int, float or str, optional
            If numeric, indicates the bandwidth and must be positive.
            If str, indicates the method to estimate the bandwidth and must be "taylor"
            since it is the only option supported so far. Defaults to "taylor".
        bw_fct: float, optional
            A value that multiplies `bw` which enables tuning smoothness by hand. Must be positive.
            Values above 1 decrease smoothness while values below 1 decrease it.
            Defaults to 1 (no modification).
        custom_lims: list or tuple, optional
            A list or tuple of length 2 indicating custom bounds for the range of `x`.
            Defaults to None which means the estimation limits are [-pi, pi].
        cumulative: bool, optional
            Whether return the PDF or the cumulative PDF. Defaults to False.
        grid_len: int, optional
            The number of intervals used to bin the data points
            i.e. the length of the grid used in the estimation. Defaults to 512.
        """
        # All values between -pi and pi
        x = self._normalize_angle(x)

        # Check `bw_fct` is numeric and positive
        if not isinstance(bw_fct, int | float | np.integer | np.floating):
            raise TypeError(f"`bw_fct` must be a positive number, not an object of {type(bw_fct)}.")

        if bw_fct <= 0:
            raise ValueError(f"`bw_fct` must be a positive number, not {bw_fct}.")

        # Determine bandwidth
        if isinstance(bw, bool):
            raise ValueError(
                "`bw` can't be of type `bool`.\nExpected a positive numeric or 'taylor'"
            )
        if isinstance(bw, int | float) and bw < 0:
            raise ValueError(f"Numeric `bw` must be positive.\nInput: {bw:.4f}.")
        if isinstance(bw, str):
            if bw == "taylor":
                bw = self.bw_taylor(x)
            else:
                raise ValueError(f"`bw` must be a positive numeric or `taylor`, not {bw}")
        bw *= bw_fct

        # Determine grid
        if custom_lims is not None:
            custom_lims = self._check_custom_lims(custom_lims, x.min(), x.max())
            grid_min = custom_lims[0]
            grid_max = custom_lims[1]
            assert grid_min >= -np.pi, "Lower limit can't be smaller than -pi"
            assert grid_max <= np.pi, "Upper limit can't be larger than pi"
        else:
            grid_min = -np.pi
            grid_max = np.pi

        bins = np.linspace(grid_min, grid_max, grid_len + 1)
        bin_counts, bin_edges = np.histogram(x, bins=bins, density=False)
        grid = 0.5 * (bin_edges[1:] + bin_edges[:-1])

        kern = self._vonmises_pdf(x=grid, mu=0, kappa=bw)
        pdf = np.fft.fftshift(np.fft.irfft(np.fft.rfft(kern) * np.fft.rfft(bin_counts)))
        pdf /= len(x)

        if cumulative:
            pdf = pdf.cumsum() / pdf.sum()

        return grid, pdf, bw

    def _kde(self, x, circular=False, grid_len=512, **kwargs):
        x = x.flatten()
        x = x[np.isfinite(x)]
        if x.size == 0 or np.all(x == x[0]):
            warnings.warn("Your data appears to have a single value or no finite values")

            return np.zeros(grid_len), np.full(grid_len, np.nan), np.nan

        if circular:
            if circular == "degrees":
                x = np.radians(x)
            kde_fun = self.kde_circular
        else:
            kde_fun = self.kde_linear

        return kde_fun(x, grid_len=grid_len, **kwargs)

    def _fast_kde_2d(self, x, y, gridsize=(128, 128), circular=False):
        """
        2D fft-based Gaussian kernel density estimate (KDE).

        The code was adapted from https://github.com/mfouesneau/faststats

        Parameters
        ----------
        x : Numpy array or list
        y : Numpy array or list
        gridsize : tuple
            Number of points used to discretize data. Use powers of 2 for fft optimization
        circular: bool
            If True use circular boundaries. Defaults to False

        Returns
        -------
        grid: A gridded 2D KDE of the input points (x, y)
        xmin: minimum value of x
        xmax: maximum value of x
        ymin: minimum value of y
        ymax: maximum value of y
        """
        x = np.asarray(x, dtype=float)
        x = x[np.isfinite(x)]
        y = np.asarray(y, dtype=float)
        y = y[np.isfinite(y)]

        xmin, xmax = x.min(), x.max()
        ymin, ymax = y.min(), y.max()

        len_x = len(x)
        weights = np.ones(len_x)
        n_x, n_y = gridsize

        d_x = (xmax - xmin) / (n_x - 1)
        d_y = (ymax - ymin) / (n_y - 1)

        xyi = np.vstack((x, y)).T
        xyi -= [xmin, ymin]
        xyi /= [d_x, d_y]
        xyi = np.floor(xyi, xyi).T

        scotts_factor = len_x ** (-1 / 6)
        cov = np.cov(xyi, ddof=1)
        std_devs = np.diag(cov) ** 0.5
        kern_nx, kern_ny = np.round(scotts_factor * 2 * np.pi * std_devs)

        inv_cov = np.linalg.inv(cov * scotts_factor**2)

        x_x = np.arange(kern_nx) - kern_nx / 2
        y_y = np.arange(kern_ny) - kern_ny / 2
        x_x, y_y = np.meshgrid(x_x, y_y)

        kernel = np.vstack((x_x.flatten(), y_y.flatten()))
        kernel = np.dot(inv_cov, kernel) * kernel
        kernel = np.exp(-kernel.sum(axis=0) / 2)
        kernel = kernel.reshape((int(kern_ny), int(kern_nx)))

        boundary = "wrap" if circular else "symm"

        grid = coo_matrix((weights, xyi), shape=(n_x, n_y)).toarray()
        grid = convolve2d(grid, kernel, mode="same", boundary=boundary)

        norm_factor = np.linalg.det(2 * np.pi * cov * scotts_factor**2)
        norm_factor = len_x * d_x * d_y * norm_factor**0.5

        grid /= norm_factor

        return grid, xmin, xmax, ymin, ymax

    def _find_hdi_contours(self, density, hdi_probs):
        """
        Find contours enclosing regions of highest posterior density.

        Parameters
        ----------
        density : array-like
            A 2D KDE on a grid with cells of equal area.
        hdi_probs : array-like
            An array of highest density interval confidence probabilities.

        Returns
        -------
        contour_levels : array
            The contour levels corresponding to the given HDI probabilities.
        """
        # Using the algorithm from corner.py
        sorted_density = np.sort(density, axis=None)[::-1]
        sm = sorted_density.cumsum()
        sm /= sm[-1]

        contours = np.empty_like(hdi_probs)
        for idx, hdi_prob in enumerate(hdi_probs):
            try:
                contours[idx] = sorted_density[sm <= hdi_prob][-1]
            except IndexError:
                contours[idx] = sorted_density[0]

        return contours



"""Core stats functions.

Functions that are needed by multiple "organization classes"
should go here. e.g. fft is used for kde bandwidth estimation and for ess.
"""

import warnings

import numpy as np
from scipy.fftpack import next_fast_len
from scipy.interpolate import CubicSpline
from scipy.stats import circmean


class _CoreBase:
    def fft(self, x):  # pylint: disable=no-self-use
        return np.fft.fft(x)

    def rfft(self, ary, n, axis=-1):  # pylint: disable=no-self-use
        return np.fft.rfft(ary, n=n, axis=axis)

    def irfft(self, ary, n, axis=-1):  # pylint: disable=no-self-use
        return np.fft.irfft(ary, n=n, axis=axis)

    def autocov(self, ary, axis=-1):
        """Compute autocovariance estimates for every lag for the input array.

        Parameters
        ----------
        ary : Numpy array
            An array containing MCMC samples

        Returns
        -------
        acov: Numpy array same size as the input array
        """
        axis = axis if axis > 0 else len(ary.shape) + axis
        n = ary.shape[axis]
        m = next_fast_len(2 * n)

        ary = ary - ary.mean(axis, keepdims=True)

        # added to silence tuple warning for a submodule
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            ifft_ary = self.rfft(ary, n=m, axis=axis)
            ifft_ary *= np.conjugate(ifft_ary)

            shape = tuple(
                slice(None) if dim_len != axis else slice(0, n)
                for dim_len, _ in enumerate(ary.shape)
            )
            cov = self.irfft(ifft_ary, n=m, axis=axis)[shape]
            cov /= n

        return cov

    def autocorr(self, ary, axis=-1):
        """Compute autocorrelation using FFT for every lag for the input array.

        See https://en.wikipedia.org/wiki/autocorrelation#Efficient_computation

        Parameters
        ----------
        ary : Numpy array
            An array containing MCMC samples

        Returns
        -------
        acorr: Numpy array same size as the input array
        """
        corr = self.autocov(ary, axis=axis)
        axis = axis = axis if axis > 0 else len(corr.shape) + axis
        norm = tuple(
            slice(None, None) if dim != axis else slice(None, 1) for dim, _ in enumerate(corr.shape)
        )
        with np.errstate(invalid="ignore"):
            corr /= corr[norm]
        return corr

    def circular_mean(self, ary):  # pylint: disable=no-self-use
        """Compute mean of circular variable measured in radians.

        The result is between -pi and pi.
        """
        return circmean(ary, high=np.pi, low=-np.pi)

    def _circular_standardize(self, ary):  # pylint: disable=no-self-use
        """Standardize circular data to the interval [-pi, pi]."""
        return np.mod(ary + np.pi, 2 * np.pi) - np.pi

    def quantile(self, ary, quantile, **kwargs):  # pylint: disable=no-self-use
        """Compute the quantile of an array of samples.

        Notes
        -----
        Should default to NumPy's "linear" method, which corresponds to
        type 7 from H&F [1]_.

        References
        ----------
        .. [1] R. J. Hyndman and Y. Fan,
        "Sample quantiles in statistical packages,"
        The American Statistician, 50(4), pp. 361-365, 1996
        """
        skipna = kwargs.pop("skipna", False)
        if skipna:
            result = np.nanquantile(ary, quantile, **kwargs)
        else:
            result = np.quantile(ary, quantile, **kwargs)
        if np.ndim(quantile) == 0:
            return result
        return np.moveaxis(result, 0, -1)

    def eti(self, ary, prob, **kwargs):
        edge_prob = (1 - prob) / 2
        return self.quantile(ary, [edge_prob, 1 - edge_prob], **kwargs)

    def _float_rankdata(self, ary):  # pylint: disable=no-self-use
        """Compute ranks on continuous data, assuming there are no ties.

        Notes
        -----
        :func:`scipy.stats.rankdata` is focused on discrete data and different ways
        to resolve ties. However, our most common use is converting all data to continuous
        to get rid of the ties, the call rankdata which is not very efficient nor
        numba compatible.
        """
        ranks = np.empty(len(ary), dtype=int)
        ranks[np.argsort(ary, axis=None)] = np.arange(1, ary.size + 1)
        return ranks

    def _compute_ranks(self, ary, relative=False):
        """Compute ranks for continuous and discrete variables."""
        ary_shape = ary.shape
        ary = ary.flatten()
        if ary.dtype.kind == "i":
            min_ary, max_ary = min(ary), max(ary)
            x = np.linspace(min_ary, max_ary, len(ary))
            csi = CubicSpline(x, ary)
            ary = csi(np.linspace(min_ary + 0.001, max_ary - 0.001, len(ary))).reshape(ary_shape)
        out = self._float_rankdata(ary).reshape(ary_shape)
        if relative:
            return out / out.size
        return out

    def _get_bininfo(self, values, bins="arviz"):
        dtype = values.dtype.kind

        if isinstance(bins, str) and bins != "arviz":
            bins = np.histogram_bin_edges(values, bins=bins)

        if isinstance(bins, np.ndarray):
            return bins[0], bins[-1], bins[1] - bins[0]

        if dtype == "i":
            x_min = values.min().astype(int)
            x_max = values.max().astype(int)
        else:
            x_min = values.min().astype(float)
            x_max = values.max().astype(float)

        if isinstance(bins, int):
            width = (x_max - x_min) / bins
            if dtype == "i":
                width = max(1, width)
            return x_min, x_max, width

        # Sturges histogram bin estimator
        width_sturges = (x_max - x_min) / (np.log2(values.size) + 1)

        # The Freedman-Diaconis histogram bin estimator.
        iqr = np.subtract(*self.quantile(values, [0.75, 0.25]))  # pylint: disable=assignment-from-no-return
        width_fd = 2 * iqr * values.size ** (-1 / 3)

        if dtype == "i":
            width = np.round(np.max([1, width_sturges, width_fd])).astype(int)
        else:
            width = np.max([width_sturges, width_fd])

        return x_min, x_max, width

    def _get_bins(self, values, bins="arviz"):
        """
        Automatically compute the number of bins for histograms.

        Parameters
        ----------
        values : array_like
        bins : int, str or array_like, default "arviz"
            If `bins` "arviz", use ArviZ default rule (explained in detail in notes),
            if it is a different string it is passed to :func:`numpy.histogram_bin_edges`.
            If `bins` is an integer it is interpreted as the number of bins, however,
            if `values` holds discrete data, there is an extra check to prevent
            the width of the bins to be smaller than ``1``.
            If it is an array it is returned as it.

        Returns
        -------
        array with the bins

        Notes
        -----
        Computes the width of the bins by taking the maximum of the Sturges and
        the Freedman-Diaconis estimators.
        According to numpy `np.histogram` this provides good all around performance.

        The Sturges is a very simplistic estimator based on the assumption of normality of the data.
        This estimator has poor performance for non-normal data, which becomes especially obvious
        for large data sets. The estimate depends only on size of the data.

        The Freedman-Diaconis rule uses interquartile range (IQR) to estimate the binwidth.
        It is considered a robust version of the Scott rule as the IQR is less affected by outliers
        than the standard deviation. However, the IQR depends on fewer points than the standard
        deviation, so it is less accurate, especially for long tailed distributions.
        """
        dtype = values.dtype.kind

        x_min, x_max, width = self._get_bininfo(values, bins)

        if dtype == "i":
            bins = np.arange(x_min, x_max + width + 1, width)
        else:
            if np.isclose(x_min, x_max):
                width = 1e-3
            bins = np.arange(x_min, x_max + width, width)

        return bins

    # pylint: disable=redefined-builtin
    def _histogram(self, ary, bins=None, range=None, weights=None, density=None):
        if bins is None:
            bins = self._get_bins(ary)
        return np.histogram(ary, bins=bins, range=range, weights=weights, density=density)

    def _hdi_linear_nearest_common(self, ary, prob):  # pylint: disable=no-self-use
        n = len(ary)

        ary = np.sort(ary)
        interval_idx_inc = int(np.floor(prob * n))
        n_intervals = n - interval_idx_inc
        interval_width = np.subtract(ary[interval_idx_inc:], ary[:n_intervals], dtype=np.float64)

        if len(interval_width) == 0:
            raise ValueError("Too few elements for interval calculation. ")

        min_idx = np.argmin(interval_width)
        hdi_interval = ary[[min_idx, min_idx + interval_idx_inc]]

        return hdi_interval

    def _hdi_nearest(self, ary, prob, circular, skipna):
        """Compute HDI over the flattened array as closest samples that contain the given prob."""
        ary = ary.flatten()
        if skipna:
            nans = np.isnan(ary)
            if not nans.all():
                ary = ary[~nans]

        if circular:
            mean = self.circular_mean(ary)
            ary = self._circular_standardize(ary - mean)

        hdi_interval = self._hdi_linear_nearest_common(ary, prob)

        if circular:
            hdi_interval = self._circular_standardize(hdi_interval + mean)

        return hdi_interval

    def _hdi_multimodal_continuous(
        self, ary, prob, skipna, max_modes, circular, from_sample=False, **kwargs
    ):
        """Compute HDI if the distribution is multimodal."""
        ary = ary.flatten()
        if skipna:
            ary = ary[~np.isnan(ary)]

        bins, density, _ = self.kde(ary, circular=circular, **kwargs)
        if from_sample:
            ary_density = np.interp(ary, bins, density)
            hdi_intervals, interval_probs = self._hdi_from_point_densities(
                ary, ary_density, prob, circular
            )
        else:
            dx = (bins[-1] - bins[0]) / (len(bins) - 1)
            bin_probs = density * dx

            hdi_intervals, interval_probs = self._hdi_from_bin_probabilities(
                bins, bin_probs, prob, circular, dx
            )

        return self._pad_hdi_to_maxmodes(hdi_intervals, interval_probs, max_modes)

    def _hdi_multimodal_discrete(self, ary, prob, max_modes, bins=None):
        """Compute HDI if the distribution is multimodal."""
        ary = ary.flatten()

        if bins is None:
            bins, counts = np.unique(ary, return_counts=True)
            bin_probs = counts / len(ary)
            dx = 1
        else:
            counts, edges = self._histogram(ary, bins=bins)
            bins = 0.5 * (edges[1:] + edges[:-1])
            bin_probs = counts / counts.sum()
            dx = bins[1] - bins[0]

        hdi_intervals, interval_probs = self._hdi_from_bin_probabilities(
            bins, bin_probs, prob, False, dx
        )

        return self._pad_hdi_to_maxmodes(hdi_intervals, interval_probs, max_modes)

    def _hdi_from_point_densities(self, points, densities, prob, circular):
        if circular:
            points = self._circular_standardize(points)

        sorted_idx = np.argsort(points)
        points = points[sorted_idx]
        densities = densities[sorted_idx]

        # find idx of points in the interval
        interval_size = int(np.ceil(prob * len(points)))
        sorted_idx = np.argsort(densities)[::-1]
        idx_in_interval = sorted_idx[:interval_size]
        idx_in_interval.sort()

        # find idx of interval bounds
        probs_in_interval = np.full(idx_in_interval.shape, 1 / len(points))
        interval_bounds_idx, interval_probs = self._interval_points_to_bounds(
            idx_in_interval, probs_in_interval, 1, circular, period=len(points)
        )

        return points[interval_bounds_idx], interval_probs

    def _hdi_from_bin_probabilities(self, bins, bin_probs, prob, circular, dx):
        if circular:
            bins = self._circular_standardize(bins)
            sorted_idx = np.argsort(bins)
            bins = bins[sorted_idx]
            bin_probs = bin_probs[sorted_idx]

        # find idx of bins in the interval
        sorted_idx = np.argsort(bin_probs)[::-1]
        cum_probs = bin_probs[sorted_idx].cumsum()
        interval_size = np.searchsorted(cum_probs, prob, side="left") + 1
        idx_in_interval = sorted_idx[:interval_size]
        idx_in_interval.sort()

        # get points in intervals
        intervals = bins[idx_in_interval]
        probs_in_interval = bin_probs[idx_in_interval]

        return self._interval_points_to_bounds(intervals, probs_in_interval, dx, circular)

    def _interval_points_to_bounds(self, points, probs, dx, circular, period=2 * np.pi):  # pylint: disable=no-self-use
        cum_probs = probs.cumsum()

        is_bound = np.diff(points) > dx * 1.01
        is_lower_bound = np.insert(is_bound, 0, True)
        is_upper_bound = np.append(is_bound, True)
        interval_bounds = np.column_stack([points[is_lower_bound], points[is_upper_bound]])
        interval_probs = (
            cum_probs[is_upper_bound] - cum_probs[is_lower_bound] + probs[is_lower_bound]
        )

        if (
            circular
            and np.mod(dx * 1.01 + interval_bounds[-1, -1] - interval_bounds[0, 0], period)
            <= dx * 1.01
        ):
            interval_bounds[-1, 1] = interval_bounds[0, 1]
            interval_bounds = interval_bounds[1:, :]
            interval_probs[-1] += interval_probs[0]
            interval_probs = interval_probs[1:]

        return interval_bounds, interval_probs

    def _pad_hdi_to_maxmodes(self, hdi_intervals, interval_probs, max_modes):  # pylint: disable=no-self-use
        if hdi_intervals.shape[0] > max_modes:
            warnings.warn(
                f"found more modes than {max_modes}, returning only the {max_modes} highest "
                "probability modes"
            )
            hdi_intervals = hdi_intervals[np.argsort(interval_probs)[::-1][:max_modes], :]
        elif hdi_intervals.shape[0] < max_modes:
            hdi_intervals = np.vstack(
                [hdi_intervals, np.full((max_modes - hdi_intervals.shape[0], 2), np.nan)]
            )
        return hdi_intervals



"""ArviZ stats general utility functions."""

import warnings
from collections.abc import Hashable
from dataclasses import dataclass
from importlib import import_module

import numpy as np
from arviz_base import rcParams
from xarray import DataArray

__all__ = ["ELPDData", "get_function", "get_log_likelihood"]


def get_function(func_name):
    """Get a function from arviz_stats.

    Attempts to import the provided function from dataarray class in the module indicated
    in the rcParam ``stats.module``, and if it fails, it imports it from ``arviz_stats.base``.

    Parameters
    ----------
    func_name : str
        Name of the function to be imported and returned

    Returns
    -------
    callable
    """
    module_name = rcParams["stats.module"]
    if isinstance(module_name, str):
        preferred_module = import_module(f"arviz_stats.{module_name}")
    else:
        preferred_module = module_name
    if hasattr(preferred_module, "dataarray_stats"):
        preferred_module = preferred_module.dataarray_stats
    if not hasattr(preferred_module, func_name):
        raise KeyError(f"Requested function '{func_name}' is not available in '{preferred_module}'")
    return getattr(preferred_module, func_name)


def get_array_function(func_name):
    """Get a function from arviz_stats' array layer.

    Attempts to import the provided function from array class in the module indicated
    in the rcParam ``stats.module``, and if it fails, it imports it from ``arviz_stats.base``.

    Parameters
    ----------
    func_name : str
        Name of the function to be imported and returned

    Returns
    -------
    callable
    """
    module_name = rcParams["stats.module"]
    if isinstance(module_name, str):
        preferred_module = import_module(f"arviz_stats.{module_name}")
    else:
        preferred_module = module_name
    if hasattr(preferred_module, "array_stats"):
        preferred_module = preferred_module.array_stats
    if not hasattr(preferred_module, func_name):
        raise KeyError(f"Requested function '{func_name}' is not available in '{preferred_module}'")
    return getattr(preferred_module, func_name)


def get_log_likelihood(idata, var_name=None):
    """Retrieve the log likelihood dataarray of a given variable."""
    if (
        not hasattr(idata, "log_likelihood")
        and hasattr(idata, "sample_stats")
        and hasattr(idata.sample_stats, "log_likelihood")
    ):
        warnings.warn(
            "Storing the log_likelihood in sample_stats groups has been deprecated",
            DeprecationWarning,
        )
        return idata.sample_stats.log_likelihood
    if not hasattr(idata, "log_likelihood"):
        raise TypeError("log likelihood not found in inference data object")
    if var_name is None:
        var_names = list(idata.log_likelihood.data_vars)
        if len(var_names) > 1:
            raise TypeError(
                f"Found several log likelihood arrays {var_names}, var_name cannot be None"
            )
        return idata.log_likelihood[var_names[0]]
    try:
        log_likelihood = idata.log_likelihood[var_name]
    except KeyError as err:
        raise TypeError(f"No log likelihood data named {var_name} found") from err
    return log_likelihood


# get_log_likelihood and get_log_prior functions should be somewhere else
def get_log_likelihood_dataset(idata, var_names=None):
    """Retrieve the log likelihood dataarray of a given variable."""
    if (
        not hasattr(idata, "log_likelihood")
        and hasattr(idata, "sample_stats")
        and hasattr(idata.sample_stats, "log_likelihood")
    ):
        warnings.warn(
            "Storing the log_likelihood in sample_stats groups has been deprecated",
            DeprecationWarning,
        )
        log_lik_ds = idata.sample_stats.ds[["log_likelihood"]]
    if not hasattr(idata, "log_likelihood"):
        raise TypeError("log likelihood not found in inference data object")
    log_lik_ds = idata.log_likelihood.ds
    if var_names is None:
        return log_lik_ds
    if isinstance(var_names, Hashable):
        return log_lik_ds[[var_names]]
    return log_lik_ds[var_names]


def get_log_likelihood_dataarray(data, var_name=None):
    log_lik_ds = get_log_likelihood_dataset(data)
    if var_name is None:
        var_names = list(log_lik_ds.data_vars)
        if len(var_names) > 1:
            raise TypeError(
                f"Found several log likelihood arrays {var_names}, var_name cannot be None"
            )
        return log_lik_ds[var_names[0]]

    try:
        log_likelihood = log_lik_ds[var_name]
    except KeyError as err:
        raise TypeError(f"No log likelihood data named {var_name} found") from err
    return log_likelihood


def get_log_prior(idata, var_names=None):
    """Retrieve the log prior dataarray of a given variable."""
    if not hasattr(idata, "log_prior"):
        raise TypeError("log prior not found in inference data object")
    if var_names is None:
        return idata.log_prior.ds
    if isinstance(var_names, Hashable):
        return idata.log_prior.ds[[var_names]]
    return idata.log_prior.ds[var_names]


BASE_FMT = """Computed from {{n_samples}} posterior samples and \
{{n_points}} observations log-likelihood matrix.

{{0:{0}}} Estimate       SE
{{scale}}_{{kind}} {{ic_value:8.2f}}  {{ic_se:7.2f}}
p_{{kind:{1}}} {{p_value:8.2f}}        -"""
POINTWISE_LOO_FMT = """------

Pareto k diagnostic values:
                         {{0:>{0}}} {{1:>6}}
(-Inf, {{8:.2f}}]   (good)     {{2:{0}d}} {{5:6.1f}}%
   ({{8:.2f}}, 1]   (bad)      {{3:{0}d}} {{6:6.1f}}%
    (1, Inf)   (very bad) {{4:{0}d}} {{7:6.1f}}%
"""
SCALE_DICT = {"deviance": "deviance", "log": "elpd", "negative_log": "-elpd"}


@dataclass
class ELPDData:  # pylint: disable=too-many-ancestors, too-many-instance-attributes
    """Class to contain the data from elpd information criterion like waic or loo."""

    kind: str
    elpd: float
    se: float
    p: float
    n_samples: int
    n_data_points: int
    scale: str
    warning: bool
    good_k: float
    elpd_i: DataArray = None
    pareto_k: DataArray = None

    def __str__(self):
        """Print elpd data in a user friendly way."""
        kind = self.kind
        scale_str = SCALE_DICT[self["scale"]]
        padding = len(scale_str) + len(kind) + 1
        base = BASE_FMT.format(padding, padding - 2)
        base = base.format(
            "",
            kind=kind,
            scale=scale_str,
            n_samples=self.n_samples,
            n_points=self.n_data_points,
            ic_value=self.elpd,
            ic_se=self.se,
            p_value=self.p,
        )

        if self.warning:
            base += "\n\nThere has been a warning during the calculation. Please check the results."

        if kind == "loo" and self.pareto_k is not None:
            bins = bins = np.asarray([-np.inf, self.good_k, 1, np.inf])
            counts, *_ = np.histogram(self.pareto_k, bins=bins, density=False)
            extended = POINTWISE_LOO_FMT.format(max(4, len(str(np.max(counts)))))
            extended = extended.format(
                "Count",
                "Pct.",
                *[*counts, *(counts / np.sum(counts) * 100)],
                self.good_k,
            )
            base = "\n".join([base, extended])
        return base

    def __repr__(self):
        """Alias to ``__str__``."""
        return self.__str__()

    def __getitem__(self, key):
        """Define getitem magic method."""
        return getattr(self, key)

    def __setitem__(self, key, item):
        """Define setitem magic method."""
        setattr(self, key, item)



"""Class with dataarray numbified functions."""

from arviz_stats.base.dataarray import BaseDataArray
from arviz_stats.numba.array import array_stats


class NumbaDataArray(BaseDataArray):
    """DataArray compatible functions that use numba."""


dataarray_stats = NumbaDataArray(array_class=array_stats)



# pylint: disable=too-many-lines, too-many-function-args, redefined-outer-name
"""Diagnostic functions for ArviZ."""

import warnings
from collections.abc import Sequence

import numpy as np
from scipy import stats
from scipy.special import logsumexp

from arviz_stats.base.core import _CoreBase
from arviz_stats.base.stats_utils import not_valid as _not_valid


class _DiagnosticsBase(_CoreBase):
    """Class with numpy.scipy only diagnostic related functions."""

    def _backtransform_ranks(self, arr, c=3 / 8):  # pylint: disable=invalid-name, no-self-use
        """Backtransformation of ranks.

        Parameters
        ----------
        arr : np.ndarray
            Ranks array
        c : float
            Fractional offset. Defaults to c = 3/8 as recommended by Blom (1958).

        Returns
        -------
        np.ndarray

        References
        ----------
        Blom, G. (1958). Statistical Estimates and Transformed Beta-Variables. Wiley; New York.
        """
        arr = np.asarray(arr)
        size = arr.size
        return (arr - c) / (size - 2 * c + 1)

    def _z_scale(self, ary):
        """Calculate z_scale.

        Parameters
        ----------
        ary : np.ndarray

        Returns
        -------
        np.ndarray
        """
        ary = np.asarray(ary)
        rank = stats.rankdata(ary, method="average", nan_policy="omit")
        rank = self._backtransform_ranks(rank)
        z = stats.norm.ppf(rank)
        z = z.reshape(ary.shape)
        return z

    def _split_chains(self, ary):  # pylint: disable=no-self-use
        """Split and stack chains."""
        ary = np.asarray(ary)
        if len(ary.shape) <= 1:
            ary = np.atleast_2d(ary)
        _, n_draw = ary.shape
        half = n_draw // 2
        return np.vstack((ary[:, :half], ary[:, -half:]))

    def _z_fold(self, ary):
        """Fold and z-scale values."""
        ary = np.asarray(ary)
        ary = abs(ary - np.median(ary))
        ary = self._z_scale(ary)
        return ary

    def _rhat(self, ary):  # pylint: disable=no-self-use
        """Compute the rhat for a 2d array."""
        ary = np.asarray(ary, dtype=float)
        _, num_samples = ary.shape

        # Calculate chain mean
        chain_mean = np.mean(ary, axis=1)
        # Calculate chain variance
        chain_var = np.var(ary, axis=1, ddof=1)
        # Calculate between-chain variance
        between_chain_variance = num_samples * np.var(chain_mean, axis=None, ddof=1)
        # Calculate within-chain variance
        within_chain_variance = np.mean(chain_var)
        # Estimate of marginal posterior variance
        rhat_value = np.sqrt(
            (between_chain_variance / within_chain_variance + num_samples - 1) / (num_samples)
        )
        return rhat_value

    def _rhat_rank(self, ary):
        """Compute the rank normalized rhat for 2d array.

        Computation follows https://arxiv.org/abs/1903.08008
        """
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 2}):
            return np.nan
        split_ary = self._split_chains(ary)
        rhat_bulk = self._rhat(self._z_scale(split_ary))

        split_ary_folded = abs(split_ary - np.median(split_ary))
        rhat_tail = self._rhat(self._z_scale(split_ary_folded))

        rhat_rank = max(rhat_bulk, rhat_tail)
        return rhat_rank

    def _rhat_folded(self, ary):
        """Calculate split-Rhat for folded z-values."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 2}):
            return np.nan
        ary = self._z_fold(self._split_chains(ary))
        return self._rhat(ary)

    def _rhat_z_scale(self, ary):
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 2}):
            return np.nan
        return self._rhat(self._z_scale(self._split_chains(ary)))

    def _rhat_split(self, ary):
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 2}):
            return np.nan
        return self._rhat(self._split_chains(ary))

    def _rhat_identity(self, ary):
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 2}):
            return np.nan
        return self._rhat(ary)

    def _ess(self, ary, relative=False):
        """Compute the effective sample size for a 2D array."""
        ary = np.asarray(ary, dtype=float)
        if (np.max(ary) - np.min(ary)) < np.finfo(float).resolution:  # pylint: disable=no-member
            return ary.size
        n_chain, n_draw = ary.shape
        acov = self.autocov(ary, axis=1)
        chain_mean = ary.mean(axis=1)
        mean_var = np.mean(acov[:, 0]) * n_draw / (n_draw - 1.0)
        var_plus = mean_var * (n_draw - 1.0) / n_draw
        if n_chain > 1:
            var_plus += np.var(chain_mean, axis=None, ddof=1)

        rho_hat_t = np.zeros(n_draw)
        rho_hat_even = 1.0
        rho_hat_t[0] = rho_hat_even
        rho_hat_odd = 1.0 - (mean_var - np.mean(acov[:, 1])) / var_plus
        rho_hat_t[1] = rho_hat_odd

        # Geyer's initial positive sequence
        t = 1
        while t < (n_draw - 3) and (rho_hat_even + rho_hat_odd) > 0.0:
            rho_hat_even = 1.0 - (mean_var - np.mean(acov[:, t + 1])) / var_plus
            rho_hat_odd = 1.0 - (mean_var - np.mean(acov[:, t + 2])) / var_plus
            if (rho_hat_even + rho_hat_odd) >= 0:
                rho_hat_t[t + 1] = rho_hat_even
                rho_hat_t[t + 2] = rho_hat_odd
            t += 2

        max_t = t - 2
        # improve estimation
        if rho_hat_even > 0:
            rho_hat_t[max_t + 1] = rho_hat_even
        # Geyer's initial monotone sequence
        t = 1
        while t <= max_t - 2:
            if (rho_hat_t[t + 1] + rho_hat_t[t + 2]) > (rho_hat_t[t - 1] + rho_hat_t[t]):
                rho_hat_t[t + 1] = (rho_hat_t[t - 1] + rho_hat_t[t]) / 2.0
                rho_hat_t[t + 2] = rho_hat_t[t + 1]
            t += 2

        ess = n_chain * n_draw
        tau_hat = (
            -1.0 + 2.0 * np.sum(rho_hat_t[: max_t + 1]) + np.sum(rho_hat_t[max_t + 1 : max_t + 2])
        )
        tau_hat = max(tau_hat, 1 / np.log10(ess))
        ess = (1 if relative else ess) / tau_hat
        if np.isnan(rho_hat_t).any():
            ess = np.nan
        return ess

    def _ess_bulk(self, ary, relative=False):
        """Compute the effective sample size for the bulk."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        z_scaled = self._z_scale(self._split_chains(ary))
        ess_bulk = self._ess(z_scaled, relative=relative)
        return ess_bulk

    def _ess_tail(self, ary, prob=None, relative=False):
        """Compute the effective sample size for the tail.

        If `prob` defined, ess = min(qess(prob), qess(1-prob))
        """
        if prob is None:
            prob = (0.05, 0.95)
        elif not isinstance(prob, Sequence):
            prob = (prob, 1 - prob)

        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan

        prob_low, prob_high = prob
        quantile_low_ess = self._ess_quantile(ary, prob_low, relative=relative)
        quantile_high_ess = self._ess_quantile(ary, prob_high, relative=relative)
        return min(quantile_low_ess, quantile_high_ess)

    def _ess_mean(self, ary, relative=False):
        """Compute the effective sample size for the mean."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        return self._ess(self._split_chains(ary), relative=relative)

    def _ess_sd(self, ary, relative=False):
        """Compute the effective sample size for the sd."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        ary = self._split_chains(ary)
        return min(self._ess(ary, relative=relative), self._ess(ary**2, relative=relative))

    def _ess_quantile(self, ary, prob, relative=False):
        """Compute the effective sample size for the specific residual."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        if prob is None:
            raise TypeError("Prob not defined.")
        quantile = self.quantile(ary, prob)
        iquantile = ary <= quantile
        return self._ess(self._split_chains(iquantile), relative=relative)

    def _ess_local(self, ary, prob, relative=False):
        """Compute the effective sample size for the specific residual."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        if prob is None:
            raise TypeError("Prob not defined.")
        if len(prob) != 2:
            raise ValueError("Prob argument in ess local must be upper and lower bound")
        quantile = self.quantile(ary, prob)
        iquantile = (quantile[0] <= ary) & (ary <= quantile[1])
        return self._ess(self._split_chains(iquantile), relative=relative)

    def _ess_z_scale(self, ary, relative=False):
        """Calculate ess for z-scaLe."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        return self._ess(self._z_scale(self._split_chains(ary)), relative=relative)

    def _ess_folded(self, ary, relative=False):
        """Calculate split-ess for folded data."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        return self._ess(self._z_fold(self._split_chains(ary)), relative=relative)

    def _ess_median(self, ary, relative=False):
        """Calculate split-ess for median."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        return self._ess_quantile(ary, 0.5, relative=relative)

    def _ess_mad(self, ary, relative=False):
        """Calculate split-ess for mean absolute deviance."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        ary = abs(ary - np.median(ary))
        ary = ary <= np.median(ary)
        ary = self._z_scale(self._split_chains(ary))
        return self._ess(ary, relative=relative)

    def _ess_identity(self, ary, relative=False):
        """Calculate ess."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        return self._ess(ary, relative=relative)

    def _mcse_mean(self, ary):
        """Compute the Markov Chain mean error."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        ess = self._ess_mean(ary)
        sd = np.std(ary, ddof=1)
        mcse_mean_value = sd / np.sqrt(ess)
        return mcse_mean_value

    def _mcse_sd(self, ary):
        """Compute the Markov Chain sd error."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        ess = self._ess_sd(ary)
        sd = np.std(ary, ddof=1)
        fac_mcse_sd = np.sqrt(np.exp(1) * (1 - 1 / ess) ** (ess - 1) - 1)
        mcse_sd_value = sd * fac_mcse_sd
        return mcse_sd_value

    def _mcse_median(self, ary):
        """Compute the Markov Chain median error."""
        return self._mcse_quantile(ary, 0.5)

    def _mcse_quantile(self, ary, prob):
        """Compute the Markov Chain quantile error at quantile=prob."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        ess = self._ess_quantile(ary, prob)
        probability = [0.1586553, 0.8413447]
        with np.errstate(invalid="ignore"):
            ppf = stats.beta.ppf(probability, ess * prob + 1, ess * (1 - prob) + 1)
        sorted_ary = np.sort(ary.ravel())
        size = sorted_ary.size
        ppf_size = ppf * size - 1
        th1 = sorted_ary[int(np.floor(np.nanmax((ppf_size[0], 0))))]
        th2 = sorted_ary[int(np.ceil(np.nanmin((ppf_size[1], size - 1))))]
        return (th2 - th1) / 2

    def _pareto_min_ss(self, ary):
        """Compute the minimum effective sample size."""
        ary = np.asarray(ary)
        if _not_valid(ary, shape_kwargs={"min_draws": 4, "min_chains": 1}):
            return np.nan
        ary_flatten = ary.flatten()
        r_eff = self._ess_tail(ary, relative=True)

        _, kappa = self._pareto_khat(ary_flatten, r_eff=r_eff, tail="both", log_weights=False)

        if kappa < 1:
            return 10 ** (1 / (1 - max(0, kappa)))

        return np.inf

    def _psislw(self, ary, r_eff):
        ary_shape = ary.shape
        ary = ary.flatten()
        n_draws = len(ary)
        n_draws_tail = self._get_ps_tails(n_draws, r_eff, tail="right")
        ary, khat = self._ps_tail(
            -ary, n_draws, n_draws_tail, smooth_draws=True, tail="right", log_weights=True
        )

        return ary.reshape(ary_shape), khat

    def _pareto_khat(self, ary, r_eff=1, tail="both", log_weights=False):
        """
        Compute Pareto k-hat diagnostic.

        See details in Vehtari et al., 2024 (https://doi.org/10.48550/arXiv.1507.02646)

        Parameters
        ----------
        ary : Array
        r_eff : float, optional
            Relative efficiency. Effective sample size divided the number of samples.
        tail : srt, optional
            Which tail to fit. Can be 'right', 'left', or 'both'.
        log_weights : bool, optional
            Whether dt represents log-weights.

        Returns
        -------
        khat : float
            Pareto k-hat value.
        """
        if log_weights:
            tail = "right"

        n_draws = len(ary)

        n_draws_tail = self._get_ps_tails(n_draws, r_eff, tail=tail)

        if tail == "both":
            khat = max(
                self._ps_tail(ary, n_draws, n_draws_tail, smooth_draws=False, tail=t)[1]
                for t in ("left", "right")
            )
        else:
            ary, khat = self._ps_tail(ary, n_draws, n_draws_tail, smooth_draws=False, tail=tail)

        return ary, khat

    @staticmethod
    def _get_ps_tails(n_draws, r_eff, tail):
        if n_draws > 255:
            n_draws_tail = int(min(n_draws * 0.2, np.ceil(3 * (n_draws / r_eff) ** 0.5)))
        else:
            n_draws_tail = int(n_draws / 5)

        if tail == "both":
            half_n_draws = n_draws // 2
            if n_draws_tail > half_n_draws:
                warnings.warn(
                    "Number of tail draws cannot be more than half "
                    "the total number of draws if both tails are fit, "
                    f"changing to {half_n_draws}"
                )
                n_draws_tail = half_n_draws

            if n_draws_tail < 5:
                warnings.warn("Number of tail draws cannot be less than 5. Changing to 5")
                n_draws_tail = 5

        return n_draws_tail

    def _ps_tail(
        self, ary, n_draws, n_draws_tail, smooth_draws=False, tail="both", log_weights=False
    ):
        """
        Estimate the tail of a distribution using the Generalized Pareto Distribution.

        Parameters
        ----------
        x : array
            1D array.
        n_draws : int
            Number of draws.
        n_draws_tail : int
            Number of draws in the tail.
        smooth_draws : bool, optional
            Whether to smooth the tail.
        tail : str, optional
            Which tail to fit. Can be 'right', 'left', or 'both'.
        log_weights : bool, optional
            Whether x represents log-weights.

        Returns
        -------
        ary : array
            Array with smoothed tail values.
        k : float
            Estimated shape parameter.
        """
        if log_weights:
            ary = ary - np.max(ary)

        if tail not in ["right", "left", "both"]:
            raise ValueError('tail must be one of "right", "left", or "both"')

        tail_ids = np.arange(n_draws - n_draws_tail, n_draws, dtype=int)

        if tail == "left":
            ary = -ary

        ordered = np.argsort(ary)
        draws_tail = ary[ordered[tail_ids]]

        cutoff = ary[ordered[tail_ids[0] - 1]]  # largest value smaller than tail values

        max_tail = np.max(draws_tail)
        min_tail = np.min(draws_tail)

        if n_draws_tail >= 5:
            if abs(max_tail - min_tail) < np.finfo(float).tiny:
                raise ValueError("All tail values are the same")

            if log_weights:
                draws_tail = np.exp(draws_tail)
                cutoff = np.exp(cutoff)

            khat, sigma = self._gpdfit(draws_tail - cutoff)

            if np.isfinite(khat) and smooth_draws:
                p = np.arange(0.5, n_draws_tail) / n_draws_tail
                smoothed = self._gpinv(p, khat, sigma, cutoff)

                if log_weights:
                    smoothed = np.log(smoothed)

            else:
                smoothed = None
        else:
            raise ValueError("n_draws_tail must be at least 5")

        if smoothed is not None:
            smoothed[smoothed > max_tail] = max_tail
            ary[ordered[tail_ids]] = smoothed

        if tail == "left":
            ary = -ary

        # normalise weights
        if log_weights:
            ary -= logsumexp(ary)
        else:
            ary /= np.sum(ary)

        return ary, khat

    @staticmethod
    def _gpdfit(ary):
        """Estimate the parameters for the Generalized Pareto Distribution (GPD).

        Empirical Bayes estimate for the parameters (kappa, sigma) of the generalized Pareto
        distribution given the data.

        The fit uses a prior for kappa to stabilize estimates for very small (effective)
        sample sizes. The weakly informative prior is a Gaussian  centered at 0.5.
        See details in Vehtari et al., 2024 (https://doi.org/10.48550/arXiv.1507.02646)


        Parameters
        ----------
        ary: array
            sorted 1D data array

        Returns
        -------
        kappa: float
            estimated shape parameter
        sigma: float
            estimated scale parameter
        """
        prior_bs = 3
        prior_k = 10
        n = len(ary)
        m_est = 30 + int(n**0.5)

        b_ary = 1 - np.sqrt(m_est / (np.arange(1, m_est + 1, dtype=float) - 0.5))
        b_ary /= prior_bs * ary[int(n / 4 + 0.5) - 1]
        b_ary += 1 / ary[-1]

        k_ary = np.log1p(-b_ary[:, None] * ary).mean(axis=1)  # pylint: disable=no-member
        len_scale = n * (np.log(-(b_ary / k_ary)) - k_ary - 1)
        weights = 1 / np.exp(len_scale - len_scale[:, None]).sum(axis=1)

        # remove negligible weights
        real_idxs = weights >= 10 * np.finfo(float).eps
        if not np.all(real_idxs):
            weights = weights[real_idxs]
            b_ary = b_ary[real_idxs]
        # normalise weights
        weights /= weights.sum()

        # posterior mean for b
        b_post = np.sum(b_ary * weights)
        # estimate for k
        kappa = np.log1p(-b_post * ary).mean()  # pylint: disable=invalid-unary-operand-type,no-member
        # add prior for kappa
        sigma = -kappa / b_post
        kappa = (n * kappa + prior_k * 0.5) / (n + prior_k)

        return kappa, sigma

    @staticmethod
    def _gpinv(probs, kappa, sigma, mu):
        """Quantile function for generalized pareto distribution."""
        if sigma <= 0:
            return np.full_like(probs, np.nan)

        if kappa == 0:
            q = mu - sigma * np.log1p(-probs)
        else:
            q = mu + sigma * np.expm1(-kappa * np.log1p(-probs)) / kappa

        return q

    def _power_scale_sense(self, ary, lower_w, upper_w, lower_alpha, upper_alpha):
        """Compute power-scaling sensitivity by finite difference second derivative of CJS."""
        ary = np.ravel(ary)
        lower_w = np.ravel(lower_w)
        upper_w = np.ravel(upper_w)
        lower_cjs = max(self._cjs_dist(ary, lower_w), self._cjs_dist(-1 * ary, lower_w))
        upper_cjs = max(self._cjs_dist(ary, upper_w), self._cjs_dist(-1 * ary, upper_w))
        lower_grad = -1 * lower_cjs / np.log2(lower_alpha)
        upper_grad = upper_cjs / np.log2(upper_alpha)
        return (lower_grad + upper_grad) / 2

    def _power_scale_lw(self, ary, alpha):
        """Compute log weights for power-scaling component by alpha."""
        shape = ary.shape
        ary = np.ravel(ary)
        log_weights = (alpha - 1) * ary
        n_draws = len(log_weights)
        r_eff = self._ess_tail(ary, relative=True)
        n_draws_tail = self._get_ps_tails(n_draws, r_eff, tail="both")
        log_weights, _ = self._ps_tail(
            log_weights,
            n_draws,
            n_draws_tail,
            smooth_draws=True,
            log_weights=True,
        )

        return log_weights.reshape(shape)

    @staticmethod
    def _cjs_dist(ary, weights):
        """Calculate the cumulative Jensen-Shannon distance between original and weighted draws."""
        # sort draws and weights
        order = np.argsort(ary)
        ary = ary[order]
        weights = weights[order]

        binwidth = np.diff(ary)

        # ecdfs
        cdf_p = np.linspace(1 / len(ary), 1 - 1 / len(ary), len(ary) - 1)
        cdf_q = np.cumsum(weights / np.sum(weights))[:-1]

        # integrals of ecdfs
        cdf_p_int = np.dot(cdf_p, binwidth)
        cdf_q_int = np.dot(cdf_q, binwidth)

        # cjs calculation
        pq_numer = np.log2(cdf_p, out=np.zeros_like(cdf_p), where=cdf_p != 0)
        qp_numer = np.log2(cdf_q, out=np.zeros_like(cdf_q), where=cdf_q != 0)

        denom = 0.5 * (cdf_p + cdf_q)
        denom = np.log2(denom, out=np.zeros_like(denom), where=denom != 0)

        cjs_pq = np.sum(binwidth * (cdf_p * (pq_numer - denom))) + 0.5 / np.log(2) * (
            cdf_q_int - cdf_p_int
        )

        cjs_qp = np.sum(binwidth * (cdf_q * (qp_numer - denom))) + 0.5 / np.log(2) * (
            cdf_p_int - cdf_q_int
        )

        cjs_pq = max(0, cjs_pq)
        cjs_qp = max(0, cjs_qp)

        bound = cdf_p_int + cdf_q_int

        return np.sqrt((cjs_pq + cjs_qp) / bound)



"""ArviZ stats accessors."""

import warnings
from collections.abc import Hashable

import numpy as np
import xarray as xr
from arviz_base.utils import _var_names
from xarray_einstats.numba import ecdf

from arviz_stats.utils import get_function

__all__ = ["AzStatsDsAccessor", "AzStatsDaAccessor", "AzStatsDtAccessor"]


class UnsetDefault:
    pass


def update_dims(dims, da):
    """Update dims to contain only those present in da."""
    if dims is None:
        return None
    if isinstance(dims, str):
        dims = [dims]
    return [dim for dim in dims if dim in da.dims]


def update_kwargs_with_dims(da, kwargs):
    """Update kwargs dict which may have a `dims` keyword."""
    kwargs = kwargs.copy()
    if "dims" in kwargs:
        kwargs.update({"dims": update_dims(kwargs["dims"], da)})
    return kwargs


def check_var_name_subset(obj, var_name):
    if isinstance(obj, xr.Dataset):
        return obj[var_name]
    if isinstance(obj, xr.DataTree):
        return obj.ds[var_name]
    return obj


def apply_function_to_dataset(func, ds, kwargs):
    result_dicts = None

    for var_name, da in ds.items():
        updated_kwargs = update_kwargs_with_dims(da, kwargs)
        subset_kwargs = {
            key: check_var_name_subset(value, var_name) for key, value in updated_kwargs.items()
        }
        result = func(da, **subset_kwargs)

        if not isinstance(result, tuple):
            result = (result,)

        if result_dicts is None:
            result_dicts = [{} for _ in range(len(result))]

        for i, res in enumerate(result):
            result_dicts[i][var_name] = res

    datasets = tuple(xr.Dataset(res_dict) for res_dict in result_dicts)

    return datasets if len(datasets) > 1 else datasets[0]


unset = UnsetDefault()


class _BaseAccessor:
    """Base accessor class."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def _apply(self, func, **kwargs):
        raise NotImplementedError("_apply private method needs to be implemented in subclasses")

    def eti(self, prob=None, dims=None, **kwargs):
        """Compute the equal tail interval."""
        kwargs["prob"] = prob
        return self._apply("eti", dims=dims, **kwargs)

    def hdi(self, prob=None, dims=None, **kwargs):
        """Compute hdi on all variables in the dataset."""
        kwargs["prob"] = prob
        return self._apply("hdi", dims=dims, **kwargs)

    def ess(self, dims=None, method="bulk", relative=False, prob=None, **kwargs):
        """Compute the ess of all the variables in the dataset."""
        return self._apply("ess", dims=dims, method=method, relative=relative, prob=prob, **kwargs)

    def rhat(self, dims=None, method="rank", **kwargs):
        """Compute the rhat of all the variables in the dataset."""
        return self._apply("rhat", dims=dims, method=method, **kwargs)

    def mcse(self, dims=None, method="mean", prob=None, **kwargs):
        """Compute the mcse of all the variables in the dataset."""
        return self._apply("mcse", dims=dims, method=method, prob=prob, **kwargs)

    def kde(self, dims=None, **kwargs):
        """Compute the KDE for all variables in the dataset."""
        return self._apply("kde", dims=dims, **kwargs)

    def get_bins(self, dims=None, **kwargs):
        """Compute the histogram bin edges for all variables in the dataset."""
        return self._apply(get_function("get_bins"), dims=dims, **kwargs)

    def histogram(self, dims=None, **kwargs):
        """Compute the histogram for all variables in the dataset."""
        return self._apply("histogram", dims=dims, **kwargs)

    def compute_ranks(self, dims=None, relative=False, **kwargs):
        """Compute ranks for all variables in the dataset."""
        return self._apply("compute_ranks", dims=dims, relative=relative, **kwargs)

    def ecdf(self, dims=None, pit=False, **kwargs):
        """Compute the ecdf for all variables in the dataset."""
        # TODO: implement ecdf here so it doesn't depend on numba
        dt_ecdf = self._apply(ecdf, dims=dims, **kwargs).rename(ecdf_axis="plot_axis")
        if pit:
            x_values = dt_ecdf.sel(plot_axis="x")
            normalized_x = x_values / x_values.max()
            dt_ecdf = dt_ecdf.where(dt_ecdf.plot_axis != "x", normalized_x)
            # Compute difference
            diff_y = dt_ecdf.sel(plot_axis="y") - dt_ecdf.sel(plot_axis="x")
            dt_ecdf = dt_ecdf.where(dt_ecdf.plot_axis != "y", diff_y)
        return dt_ecdf

    def pareto_min_ss(self, dims=None, **kwargs):
        """Compute the min sample size for all variables in the dataset."""
        return self._apply("pareto_min_ss", dims=dims, **kwargs)

    def psislw(self, dims=None, **kwargs):
        """Pareto smoothed importance sampling."""
        return self._apply("psislw", dims=dims, **kwargs)

    def power_scale_lw(self, dims=None, **kwargs):
        """Compute log weights for power-scaling of the DataTree."""
        return self._apply("power_scale_lw", dims=dims, **kwargs)

    def power_scale_sense(self, dims=None, **kwargs):
        """Compute power-scaling sensitivity."""
        return self._apply("power_scale_sense", dims=dims, **kwargs)

    def autocorr(self, dims=None, **kwargs):
        """Compute autocorrelation for all variables in the dataset."""
        return self._apply("autocorr", dims=dims, **kwargs)


@xr.register_dataarray_accessor("azstats")
class AzStatsDaAccessor(_BaseAccessor):
    """ArviZ stats accessor class for DataArrays."""

    def _apply(self, func, **kwargs):
        """Apply function to DataArray input."""
        if isinstance(func, str):
            func = get_function(func)
        return func(self._obj, **kwargs)


@xr.register_dataset_accessor("azstats")
class AzStatsDsAccessor(_BaseAccessor):
    """ArviZ stats accessor class for Datasets.

    Notes
    -----
    Whenever "dims" indicates a set of dimensions that are to be reduced, the behaviour
    should be to reduce all present dimensions and ignore the ones not present.
    Thus, they can't use :meth:`.Dataset.map` and instead we must manually loop over variables
    in the dataset, remove elements from dims if necessary and afterwards rebuild the output
    Dataset.
    """

    @property
    def ds(self):
        """Return the underlying Dataset."""
        return self._obj

    @property
    def dataset(self):
        """Return the underlying Dataset."""
        return self._obj

    def filter_vars(self, var_names=None, filter_vars=None):
        """Filter variables in the dataset.

        Parameters
        ----------
        var_names : iterable, optional
        filter_vars : {None, "like", "regex"}, default None

        Returns
        -------
        accessor
            This method returns the accessor after filtering its underlying xarray object.
            To get the filtered dataset, use ``.ds``.
        """
        var_names = _var_names(var_names=var_names, data=self._obj, filter_vars=filter_vars)
        if var_names is not None:
            self._obj = self._obj[var_names]
        return self

    def _apply(self, func, **kwargs):
        """Apply a function to all variables subsetting dims to existing dimensions."""
        if isinstance(func, str):
            func = get_function(func)
        return apply_function_to_dataset(func, self._obj, kwargs=kwargs)

    def thin_factor(self, **kwargs):
        """Get thinning factor for all the variables in the dataset."""
        reduce_func = kwargs.get("reduce_func", "mean")
        thin_factors = self._apply("thin_factor", **kwargs).to_array()
        if reduce_func == "mean":
            return int(np.floor(thin_factors.mean().item()))
        if reduce_func == "min":
            return int(np.floor(thin_factors.min().item()))
        raise ValueError(
            f"`reduce_func` {reduce_func} not recognized. Valid values are 'mean' or 'min'"
        )

    def thin(self, dims=None, factor="auto"):
        """Perform thinning for all the variables in the dataset."""
        if factor == "auto":
            factor = self.thin_factor()
            dims = "draw"
        return self._apply("thin", dims=dims, factor=factor)


@xr.register_datatree_accessor("azstats")
class AzStatsDtAccessor(_BaseAccessor):
    """ArviZ stats accessor class for DataTrees."""

    @property
    def datatree(self):
        """Return the underlying Dataset."""
        return self._obj

    def _process_input(self, group, method, allow_non_matching=True):
        if self._obj.name == group:
            return self._obj
        if self._obj.children and group in self._obj.children:
            return self._obj[group]
        if allow_non_matching:
            warnings.warn(
                f"Computing {method} on DataTree named {self._obj.name} which doesn't match "
                f"the group argument {group}"
            )
            return self._obj
        raise ValueError(
            f"Group {group} not available in DataTree. Present groups are {self._obj.children} "
            f"and the DataTree itself is named {self._obs.name}"
        )

    def _apply(self, func, **kwargs):
        group = kwargs.pop("group", "posterior")
        hashable_group = False
        if isinstance(group, Hashable):
            group = [group]
            hashable_group = True
        out_dt = xr.DataTree.from_dict(
            {
                group_i: apply_function_to_dataset(
                    get_function(func),
                    # if group is a single str/hashable that doesn't match the group name,
                    # still allow it and apply the function to the top level of the provided input
                    self._process_input(group_i, func, allow_non_matching=hashable_group),
                    kwargs=kwargs,
                )
                for group_i in group
            }
        )
        if hashable_group:
            # if group was a string/hashable, return a datatree with a single node
            # (from the provided group) as the root of the DataTree
            return out_dt[group[0]]
        # if group was a sequence, return a DataTree with multiple groups in the 1st level,
        # as many groups as requested
        return out_dt

    def filter_vars(self, group="posterior", var_names=None, filter_vars=None):
        """Access and filter variables of the provided group."""
        if var_names is None:
            return self
        hashable_group = False
        if isinstance(group, Hashable):
            group = [group]
            hashable_group = True
        out_dt = xr.DataTree.from_dict(
            {
                group_i: self._process_input(  # pylint: disable=protected-access
                    group_i, "filter_vars", allow_non_matching=hashable_group
                )
                .dataset.azstats.filter_vars(var_names=var_names, filter_vars=filter_vars)
                ._obj
                for group_i in group
            }
        )
        if hashable_group:
            # if group was a string/hashable, return a datatree with a single node
            # (from the provided group) as the root of the DataTree
            return out_dt[group[0]].azstats
        # if group was a sequence, return a DataTree with multiple groups in the 1st level,
        # as many groups as requested
        return out_dt.azstats

    def thin_factor(self, group="posterior", **kwargs):
        """Get thinning factor for all the variables in a group of the datatree."""
        if not isinstance(group, str):
            raise ValueError("Thin factor can only be applied over a single group.")
        reduce_func = kwargs.get("reduce_func", "mean")
        thin_factors = self._apply("thin_factor", group=group, **kwargs).ds.to_array()
        if reduce_func == "mean":
            return int(np.floor(thin_factors.mean().item()))
        if reduce_func == "min":
            return int(np.floor(thin_factors.min().item()))
        raise ValueError(
            f"`reduce_func` {reduce_func} not recognized. Valid values are 'mean' or 'min'"
        )

    def thin(self, dims=None, group="posterior", **kwargs):
        """Perform thinning for all variables in a group of the DataTree."""
        if kwargs.get("factor", "auto") == "auto":
            kwargs["factor"] = self.thin_factor()
            dims = "draw"
        return self._apply("thin", dims=dims, group=group, **kwargs)



"""Functions for sampling diagnostics."""

import numpy as np
import xarray as xr
from arviz_base import convert_to_dataset

from arviz_stats.utils import get_array_function
from arviz_stats.validate import validate_dims


def ess(
    data,
    sample_dims=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    method="bulk",
    relative=False,
    prob=None,
    chain_axis=0,
    draw_axis=1,
):
    r"""Estimate the effective sample size (ess).

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset`

    sample_dims : iterable of hashable, optional
        Dimensions to be considered sample dimensions and are to be reduced.
        Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the ESS.
    var_names : str or list of str, optional
        Names of the variables for which the ess should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    method : str, default "bulk"
        Valid methods are:

        - "bulk"
        - "tail"     # prob, optional
        - "quantile" # prob
        - "mean" (old ess)
        - "sd"
        - "median"
        - "mad" (mean absolute deviance)
        - "z_scale"
        - "folded"
        - "identity"
        - "local"  # prob

    relative : bool
        Return relative ess ``ress = ess / n``
    prob : float, or tuple of two floats, optional
        Probability value for "tail", "quantile" or "local" ess functions.
    chain_axis, draw_axis : int, optional
        Integer indicators of the axis that correspond to the chain and the draw dimension.
        `chain_axis` can be ``None``.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested ESS summary of the provided input

    See Also
    --------
    arviz.rhat : Compute estimate of rank normalized split R-hat for a set of traces.
    arviz.mcse : Calculate Markov Chain Standard Error statistic.
    plot_ess : Plot quantile, local or evolution of effective sample sizes (ESS).
    arviz.summary : Create a data frame with summary statistics.

    Examples
    --------
    Calculate the effective_sample_size using the default arguments:

    .. ipython::

        In [1]: from arviz_base import load_arviz_data
           ...: import arviz_stats as azs
           ...: data = load_arviz_data('non_centered_eight')
           ...: azs.ess(data)

    Calculate ess for a subset of the variables

    .. ipython::

        In [1]: azs.ess(data, relative=True, var_names=["mu", "theta_t"])

    Calculate ess using the "tail" method, leaving the `prob` at its default value.

    .. ipython::

        In [1]: azs.ess(data, method="tail")
    """
    if isinstance(data, list | tuple | np.ndarray):
        data = np.array(data)
        return get_array_function("ess")(
            data,
            method=method,
            relative=relative,
            prob=prob,
            chain_axis=chain_axis,
            draw_axis=draw_axis,
        )

    if isinstance(data, xr.core.groupby.DataArrayGroupBy | xr.core.groupby.DatasetGroupBy):
        # Make sure the grouped dimension is added as one of the dimensions to be reduced
        sample_dims = list(set(validate_dims(sample_dims)).union(data.group1d.dims))
        return data.map(
            ess,
            sample_dims=sample_dims,
            var_names=var_names,
            coords=coords,
            method=method,
            relative=relative,
            prob=prob,
        )

    if isinstance(data, xr.DataArray):
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.ess(dims=sample_dims, method=method, relative=relative, prob=prob)

    if isinstance(data, xr.DataTree):
        data = data.azstats.filter_vars(
            group=group, var_names=var_names, filter_vars=filter_vars
        ).datatree
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.ess(
            dims=sample_dims, group=group, method=method, relative=relative, prob=prob
        )

    data = convert_to_dataset(data, group=group)

    data = data.azstats.filter_vars(var_names=var_names, filter_vars=filter_vars).dataset
    if coords is not None:
        data = data.sel(coords)

    return data.azstats.ess(dims=sample_dims, method=method, relative=relative, prob=prob)


def rhat(
    data,
    sample_dims=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    method="rank",
    chain_axis=0,
    draw_axis=1,
):
    r"""Compute estimate of rank normalized split R-hat for a set of traces.

    The rank normalized R-hat diagnostic tests for lack of convergence by comparing the variance
    between multiple chains to the variance within each chain. If convergence has been achieved,
    the between-chain and within-chain variances should be identical. To be most effective in
    detecting evidence for nonconvergence, each chain should have been initialized to starting
    values that are dispersed relative to the target distribution.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset`

        At least 2 posterior chains are needed to compute this diagnostic of one or more
        stochastic parameters.

    sample_dims : iterable of hashable, optional
        Dimensions to be considered sample dimensions and are to be reduced.
        Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the ESS.
    var_names : str or list of str, optional
        Names of the variables for which the Rhat should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    method : str, default "rank"
        Valid methods are:
        - "rank"        # recommended by Vehtari et al. (2021)
        - "split"
        - "folded"
        - "z_scale"
        - "identity"
    chain_axis, draw_axis : int, optional
        Integer indicators of the axis that correspond to the chain and the draw dimension.
        `chain_axis` can be ``None``.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested Rhat summary of the provided input


    See Also
    --------
    arviz.ess : Calculate estimate of the effective sample size (ess).
    arviz.mcse : Calculate Markov Chain Standard Error statistic.
    plot_forest : Forest plot to compare HDI intervals from a number of distributions.

    Notes
    -----
    The diagnostic is computed by:

      .. math:: \hat{R} = \sqrt{\frac{\hat{V}}{W}}

    where :math:`W` is the within-chain variance and :math:`\hat{V}` is the posterior variance
    estimate for the pooled rank-traces. This is the potential scale reduction factor, which
    converges to unity when each of the traces is a sample from the target posterior. Values
    greater than one indicate that one or more chains have not yet converged.

    Rank values are calculated over all the chains with ``scipy.stats.rankdata``.
    Each chain is split in two and normalized with the z-transform following
    Vehtari et al. (2021).

    References
    ----------
    * Vehtari et al. (2021).  Rank-normalization, folding, and
      localization: An improved Rhat for assessing convergence of
      MCMC. Bayesian analysis, 16(2):667-718.
    * Gelman et al. BDA3 (2013)
    * Brooks and Gelman (1998)
    * Gelman and Rubin (1992)

    Examples
    --------
    Calculate the R-hat using the default arguments:

    .. ipython::

        In [1]: from arviz_base import load_arviz_data
           ...: import arviz_stats as azs
           ...: data = load_arviz_data('non_centered_eight')
           ...: azs.rhat(data)

    Calculate the R-hat of some variables using the folded method:

    .. ipython::

        In [1]: azs.rhat(data, var_names=["mu", "theta_t"], method="folded")

    """
    if isinstance(data, list | tuple | np.ndarray):
        data = np.array(data)
        return get_array_function("rhat")(
            data,
            method=method,
            chain_axis=chain_axis,
            draw_axis=draw_axis,
        )

    if isinstance(data, xr.core.groupby.DataArrayGroupBy | xr.core.groupby.DatasetGroupBy):
        # Make sure the grouped dimension is added as one of the dimensions to be reduced
        sample_dims = list(set(validate_dims(sample_dims)).union(data.group1d.dims))
        return data.map(
            rhat,
            sample_dims=sample_dims,
            var_names=var_names,
            coords=coords,
            method=method,
        )

    if isinstance(data, xr.DataArray):
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.rhat(dims=sample_dims, method=method)

    if isinstance(data, xr.DataTree):
        data = data.azstats.filter_vars(
            group=group, var_names=var_names, filter_vars=filter_vars
        ).datatree
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.rhat(dims=sample_dims, group=group, method=method)

    data = convert_to_dataset(data, group=group)

    data = data.azstats.filter_vars(var_names=var_names, filter_vars=filter_vars).dataset
    if coords is not None:
        data = data.sel(coords)

    return data.azstats.rhat(dims=sample_dims, method=method)


def mcse(
    data,
    sample_dims=None,
    group="posterior",
    var_names=None,
    filter_vars=None,
    coords=None,
    method="mean",
    prob=None,
    chain_axis=0,
    draw_axis=1,
):
    """Calculate Markov Chain Standard Error statistic.

    Parameters
    ----------
    data : array-like, DataArray, Dataset, DataTree, DataArrayGroupBy, DatasetGroupBy, or idata-like
        Input data. It will have different pre-processing applied to it depending on its type:

        - array-like: call array layer within ``arviz-stats``.
        - xarray object: apply dimension aware function to all relevant subsets
        - others: passed to :func:`arviz_base.convert_to_dataset`

    sample_dims : iterable of hashable, optional
        Dimensions to be considered sample dimensions and are to be reduced.
        Default ``rcParams["data.sample_dims"]``.
    group : hashable, default "posterior"
        Group on which to compute the ESS.
    var_names : str or list of str, optional
        Names of the variables for which the mcse should be computed.
    filter_vars : {None, "like", "regex"}, default None
    coords : dict, optional
        Dictionary of dimension/index names to coordinate values defining a subset
        of the data for which to perform the computation.
    method : str, default "mean"
        Valid methods are:
        - "mean"
        - "sd"
        - "median"
        - "quantile"
    prob : float, or tuple of two floats, optional
        Probability value "quantile".
    chain_axis, draw_axis : int, optional
        Integer indicators of the axis that correspond to the chain and the draw dimension.
        `chain_axis` can be ``None``.

    Returns
    -------
    ndarray, DataArray, Dataset, DataTree
        Requested mcse summary of the provided input

    See Also
    --------
    arviz.ess : Compute autocovariance estimates for every lag for the input array.
    arviz.summary : Create a data frame with summary statistics.
    plot_mcse : Plot quantile or local Monte Carlo Standard Error.

    Examples
    --------
    Calculate the Markov Chain Standard Error using the default arguments:

    .. ipython::

        In [1]: from arviz_base import load_arviz_data
           ...: import arviz_stats as azs
           ...: data = load_arviz_data('non_centered_eight')
           ...: azs.mcse(data)

    Calculate the Markov Chain Standard Error using the quantile method:

    .. ipython::

        In [1]: azs.mcse(data, method="quantile", prob=0.7)

    """
    if isinstance(data, list | tuple | np.ndarray):
        data = np.array(data)
        return get_array_function("mcse")(
            data,
            method=method,
            prob=prob,
            chain_axis=chain_axis,
            draw_axis=draw_axis,
        )

    if isinstance(data, xr.core.groupby.DataArrayGroupBy | xr.core.groupby.DatasetGroupBy):
        # Make sure the grouped dimension is added as one of the dimensions to be reduced
        sample_dims = list(set(validate_dims(sample_dims)).union(data.group1d.dims))
        return data.map(
            mcse,
            sample_dims=sample_dims,
            var_names=var_names,
            coords=coords,
            method=method,
            prob=prob,
        )

    if isinstance(data, xr.DataArray):
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.mcse(dims=sample_dims, method=method, prob=prob)

    if isinstance(data, xr.DataTree):
        data = data.azstats.filter_vars(
            group=group, var_names=var_names, filter_vars=filter_vars
        ).datatree
        if coords is not None:
            data = data.sel(coords)
        return data.azstats.mcse(dims=sample_dims, group=group, method=method, prob=prob)

    data = convert_to_dataset(data, group=group)

    data = data.azstats.filter_vars(var_names=var_names, filter_vars=filter_vars).dataset
    if coords is not None:
        data = data.sel(coords)

    return data.azstats.mcse(dims=sample_dims, method=method, prob=prob)



"""Power-scaling sensitivity diagnostics."""

import logging
from typing import cast

import numpy as np
import pandas as pd
import xarray as xr
from arviz_base import convert_to_datatree, extract
from arviz_base.labels import BaseLabeller
from arviz_base.sel_utils import xarray_var_iter

from arviz_stats.utils import get_log_likelihood_dataset, get_log_prior
from arviz_stats.validate import validate_dims

_log = logging.getLogger(__name__)

labeller = BaseLabeller()

__all__ = ["psense", "psense_summary"]


def psense(
    data,
    var_names=None,
    filter_vars=None,
    group="prior",
    coords=None,
    sample_dims=None,
    alphas=(0.99, 1.01),
    group_var_names=None,
    group_coords=None,
):
    """
    Compute power-scaling sensitivity values.

    Parameters
    ----------
    data : DataTree or InferenceData
        Input data. It should contain the posterior and the log_likelihood and/or log_prior groups.
    var_names : list of str, optional
        Names of posterior variables to include in the power scaling sensitivity diagnostic
    filter_vars: {None, "like", "regex"}, default None
        Used for `var_names` only.
        If ``None`` (default), interpret var_names as the real variables names.
        If "like", interpret var_names as substrings of the real variables names.
        If "regex", interpret var_names as regular expressions on the real variables names.
    group : {"prior", "likelihood"}, default "prior"
        If "likelihood", the pointsize log likelihood values are retrieved
        from the ``log_likelihood`` group and added together.
        If "prior", the log prior values are retrieved from the ``log_prior`` group.
    coords : dict, optional
        Coordinates defining a subset over the posterior. Only these variables will
        be used when computing the prior sensitivity.
    sample_dims : str or sequence of hashable, optional
        Dimensions to reduce unless mapped to an aesthetic.
        Defaults to ``rcParams["data.sample_dims"]``
    alphas : tuple
        Lower and upper alpha values for gradient calculation. Defaults to (0.99, 1.01).
    group_var_names : str, optional
        Name of the prior or log likelihood variables to use
    group_coords : dict, optional
        Coordinates defining a subset over the group element for which to
        compute the prior sensitivity diagnostic.

    Returns
    -------
    xarray.DataTree
        Returns dataTree of power-scaling sensitivity diagnostic values.
        Higher sensitivity values indicate greater sensitivity.
        Prior sensitivity above 0.05 indicates informative prior.
        Likelihood sensitivity below 0.05 indicates weak or non-informative likelihood.

    Notes
    -----
    The diagnostic is computed by power-scaling either the prior or likelihood
    and determining the degree to which the posterior changes as described in [1]_.
    It uses Pareto-smoothed importance sampling to avoid refitting the model.

    References
    ----------
    .. [1] Kallioinen et al, *Detecting and diagnosing prior and likelihood sensitivity with
       power-scaling*, Stat Comput 34, 57 (2024), https://doi.org/10.1007/s11222-023-10366-5
    """
    data = convert_to_datatree(data)

    dataset = extract(
        data,
        var_names=var_names,
        filter_vars=filter_vars,
        group="posterior",
        combined=False,
        keep_dataset=True,
    )
    if coords is not None:
        dataset = dataset.sel(coords)

    lower_w, upper_w = _get_power_scale_weights(
        data,
        alphas=alphas,
        group=group,
        sample_dims=sample_dims,
        group_var_names=group_var_names,
        group_coords=group_coords,
    )

    return dataset.azstats.power_scale_sense(
        lower_w=lower_w,
        upper_w=upper_w,
        lower_alpha=alphas[0],
        upper_alpha=alphas[1],
        dims=sample_dims,
    )


def psense_summary(
    data,
    var_names=None,
    filter_vars=None,
    coords=None,
    sample_dims=None,
    threshold=0.05,
    alphas=(0.99, 1.01),
    prior_var_names=None,
    likelihood_var_names=None,
    prior_coords=None,
    likelihood_coords=None,
    round_to=3,
):
    """
    Compute the prior/likelihood sensitivity based on power-scaling perturbations.

    Parameters
    ----------
    data : DataTree or InferenceData
        Input data. It should contain the posterior and the log_likelihood and/or log_prior groups.
    var_names : list of str, optional
        Names of posterior variables to include in the power scaling sensitivity diagnostic
    filter_vars: {None, "like", "regex"}, default None
        Used for `var_names` only.
        If ``None`` (default), interpret var_names as the real variables names.
        If "like", interpret var_names as substrings of the real variables names.
        If "regex", interpret var_names as regular expressions on the real variables names.
    coords : dict, optional
        Coordinates defining a subset over the posterior. Only these variables will
        be used when computing the prior sensitivity.
    sample_dims : str or sequence of hashable, optional
        Dimensions to reduce unless mapped to an aesthetic.
        Defaults to ``rcParams["data.sample_dims"]``
    threshold : float, optional
        Threshold value to determine the sensitivity diagnosis. Default is 0.05.
    alphas : tuple
        Lower and upper alpha values for gradient calculation. Defaults to (0.99, 1.01).
    prior_var_names : str, optional
        Name of the log-prior variables to include in the power scaling sensitivity diagnostic
    likelihood_var_names : str, optional
        Name of the log-likelihood variables to include in the power scaling sensitivity diagnostic
    prior_coords : dict, optional
        Coordinates defining a subset over the group element for which to
        compute the log-prior sensitivity diagnostic
    likelihood_coords : dict, optional
        Coordinates defining a subset over the group element for which to
        compute the log-likelihood sensitivity diagnostic
    round_to : int, optional
        Number of decimal places to round the sensitivity values. Default is 3.

    Returns
    -------
    psense_df : DataFrame
        DataFrame containing the prior and likelihood sensitivity values for each variable
        in the data. And a diagnosis column with the following values:
        - "prior-data conflict" if both prior and likelihood sensitivity are above threshold
        - "strong prior / weak likelihood" if the prior sensitivity is above threshold
        and the likelihood sensitivity is below the threshold
        - "-" otherwise

    Examples
    --------
    .. ipython::

        In [1]: from arviz_base import load_arviz_data
           ...: from arviz_stats import psense_summary
           ...: rugby = load_arviz_data("rugby")
           ...: psense_summary(rugby, var_names="atts")

    Notes
    -----
    The diagnostic is computed by power-scaling either the prior or likelihood
    and determining the degree to which the posterior changes as described in [1]_.
    It uses Pareto-smoothed importance sampling to avoid refitting the model.

    References
    ----------
    .. [1] Kallioinen et al, *Detecting and diagnosing prior and likelihood sensitivity with
       power-scaling*, Stat Comput 34, 57 (2024), https://doi.org/10.1007/s11222-023-10366-5
    """
    pssdp = psense(
        data,
        var_names=var_names,
        filter_vars=filter_vars,
        group="prior",
        sample_dims=sample_dims,
        coords=coords,
        alphas=alphas,
        group_var_names=prior_var_names,
        group_coords=prior_coords,
    )
    pssdl = psense(
        data,
        var_names=var_names,
        filter_vars=filter_vars,
        group="likelihood",
        coords=coords,
        sample_dims=sample_dims,
        alphas=alphas,
        group_var_names=likelihood_var_names,
        group_coords=likelihood_coords,
    )

    joined = xr.concat([pssdp, pssdl], dim="component").assign_coords(
        component=["prior", "likelihood"]
    )

    n_vars = np.sum([joined[var].size // 2 for var in joined.data_vars])

    psense_df = pd.DataFrame(
        (np.full((cast(int, n_vars), 2), np.nan)), columns=["prior", "likelihood"]
    )

    indices = []
    for i, (var_name, sel, isel, values) in enumerate(
        xarray_var_iter(joined, skip_dims={"component"})
    ):
        psense_df.iloc[i] = values
        indices.append(labeller.make_label_flat(var_name, sel, isel))
    psense_df.index = indices

    def _diagnose(row):
        if row["prior"] >= threshold and row["likelihood"] >= threshold:
            return "potential prior-data conflict"
        if row["prior"] > threshold > row["likelihood"]:
            return "potential strong prior / weak likelihood"

        return "✓"

    psense_df["diagnosis"] = psense_df.apply(_diagnose, axis=1)

    if "potential" in "".join(psense_df["diagnosis"]):
        _log.warning(
            "We detected potential issues. For more information on how to interpret the results, "
            "please check\n"
            "https://arviz-devs.github.io/Exploratory-Analysis-of-Bayesian-Models/Chapters/"
            "Sensitivity_checks.html#interpreting-sensitivity-diagnostics-summary\n"
            "or read original paper https://doi.org/10.1007/s11222-023-10366-5"
        )

    return psense_df.round(round_to)


def power_scale_dataset(data, group, alphas, sample_dims, group_var_names, group_coords):
    """Resample posterior based on power-scaled weights.

    Parameters
    ----------
    data : DataTree or InferenceData
        Input data. It should contain the posterior and the log_likelihood and/or log_prior groups.
    group : str
        Group to resample. Either "prior" or "likelihood"
    alphas : tuple of float
        Lower and upper alpha values for power scaling.
    sample_dims : str or sequence of hashable
        Dimensions to reduce unless mapped to an aesthetic.
    group_var_names : str
        Name of the log-prior or log-likelihood variables to use.
    group_coords : dict
        Coordinates defining a subset over the group element for which to
        compute the sensitivity diagnostic.

    Returns
    -------
    DataSet with resampled data.
    """
    dt = convert_to_datatree(data)

    lower_w, upper_w = _get_power_scale_weights(
        dt,
        alphas,
        group=group,
        sample_dims=sample_dims,
        group_var_names=group_var_names,
        group_coords=group_coords,
    )

    lower_w = lower_w.values.flatten()
    upper_w = upper_w.values.flatten()
    s_size = len(lower_w)

    idxs_to_drop = sample_dims if len(sample_dims) == 1 else ["sample"] + sample_dims
    idxs_to_drop = set(idxs_to_drop).union(
        [
            idx
            for idx in dt["posterior"].xindexes
            if any(dim in dt["posterior"][idx].dims for dim in sample_dims)
        ]
    )
    resampled = [
        extract(
            dt,
            group="posterior",
            sample_dims=sample_dims,
            num_samples=s_size,
            weights=weights,
            random_seed=42,
            resampling_method="stratified",
        ).drop_indexes(idxs_to_drop)
        for weights in (lower_w, upper_w)
    ]
    resampled.insert(
        1, extract(dt, group="posterior", sample_dims=sample_dims).drop_indexes(idxs_to_drop)
    )

    return xr.concat(resampled, dim="alpha").assign_coords(alpha=[alphas[0], 1, alphas[1]])


def _get_power_scale_weights(
    dt, alphas=None, group=None, sample_dims=None, group_var_names=None, group_coords=None
):
    """Compute power scale weights."""
    sample_dims = validate_dims(sample_dims)

    if group == "likelihood":
        group_draws = get_log_likelihood_dataset(dt, var_names=group_var_names)
    elif group == "prior":
        group_draws = get_log_prior(dt, var_names=group_var_names)
    else:
        raise ValueError("Value for `group` argument not recognized")

    if group_coords is not None:
        group_draws = group_draws.sel(group_coords)
    # we stack the different variables (if any) and dimensions in each variable (if any)
    # into a flat dimension "latent-obs_var", over which we sum afterwards.
    # Consequently, after this group_draws draws is a dataarray with only sample_dims as dims
    group_draws = group_draws.to_stacked_array("latent-obs_var", sample_dims=sample_dims).sum(
        "latent-obs_var"
    )

    # calculate importance sampling weights for lower and upper alpha power-scaling
    lower_w = np.exp(group_draws.azstats.power_scale_lw(alpha=alphas[0], dims=sample_dims))
    lower_w = lower_w / lower_w.sum(sample_dims)

    upper_w = np.exp(group_draws.azstats.power_scale_lw(alpha=alphas[1], dims=sample_dims))
    upper_w = upper_w / upper_w.sum(sample_dims)

    return lower_w, upper_w
